# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2020, Vivek R <123vivekr@gmail.com>
# Copyright (c) 2022, Alex Băluț <alexandru.balut@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Pitivi's Tracker perspective."""
import json
import os
import uuid
from gettext import gettext as _
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import cairo
import numpy
from gi.repository import Gdk
from gi.repository import GES
from gi.repository import Gio
from gi.repository import GObject
from gi.repository import Gst
from gi.repository import GstController
from gi.repository import GstVideo
from gi.repository import Gtk

from pitivi.configure import get_ui_dir
from pitivi.perspective import Perspective
from pitivi.utils.loggable import Loggable
from pitivi.utils.pipeline import AssetPipeline
from pitivi.utils.pipeline import SimplePipeline
from pitivi.utils.ui import fix_infobar
from pitivi.utils.ui import NORMAL_CURSOR
from pitivi.utils.ui import PADDING
from pitivi.utils.ui import SPACING

# The meta of an Asset holding all the tracked objects data version 1.
ASSET_TRACKED_OBJECTS_META = "pitivi::tracker_data::1"

# The meta of an Effect holding the object_id of the tracked object.
EFFECT_TRACKED_OBJECT_ID_META = "pitivi:tracked_object_id"
# The meta of an Effect holding the name of the tracked object.
EFFECT_TRACKED_OBJECT_NAME_META = "pitivi:tracked_object_name"


# TODO: Replace with bisect.bisect_left when we use Python 3.10.
def bisect_left(values, val, key):
    low = 0
    high = len(values)
    while low < high:
        mid = (low + high) // 2
        if key(values[mid]) < val:
            low = mid + 1
        else:
            high = mid
    return low


class ObjectManager():
    """Manager of an Asset's tracked objects.

    Attributes:
        asset (GES.Asset): The Asset in which the objects are being tracked.
        objects (List[Tuple[int, str, str]]): The objects stored as
            (index, object_id, name) tuples.
        values (Dict[str, List[Tuple[int, Tuple[float, float, float, float]]]]):
            The tracking data for each object is kept as a list of
            (timestamp, rectangle) tuples, always ordered.
    """

    def __init__(self, asset: GES.Asset):
        self.asset: GES.Asset = asset
        # The list of objects kept in a persistent order.
        # (index, object_id, name)
        self.objects: List[Tuple[int, str, str]] = []
        # object_id -> [(timestamp, (x, y, w, h)), ...]
        self.values: Dict[str, List[Tuple[int, Tuple[float, float, float, float]]]] = {}

        dump_str = self.asset.get_string(ASSET_TRACKED_OBJECTS_META)
        if dump_str:
            data = json.loads(dump_str)
            objects, values = data
            # Convert lists back to tuples.
            self.objects = [tuple(o) for o in objects]
            self.values = {object_id: [(position, tuple(area)) for (position, area) in values_list]
                           for object_id, values_list in values.items()}

    def save(self):
        data = [self.objects, self.values]
        dump_str = json.dumps(data)
        if self.asset.check_meta_registered(ASSET_TRACKED_OBJECTS_META):
            self.asset.set_string(ASSET_TRACKED_OBJECTS_META, dump_str)
        else:
            self.asset.register_meta_string(GES.MetaFlag.READWRITE, ASSET_TRACKED_OBJECTS_META, dump_str)

    def update_object(self, object_id: str, start_pos: int, roi_data: Dict[int, Tuple[float, float, float, float]]):
        """Updates the values from the specified position to the end."""
        object_values: List[Tuple[int, Tuple[float, float, float, float]]] = self.values[object_id]
        index: int = bisect_left(object_values, start_pos, key=lambda x: x[0])
        object_values = object_values[:index]
        object_values.extend(sorted(roi_data.items()))
        self.values[object_id] = object_values

    def update_object_position(self, object_id: str, position: int, area: Tuple[float, float, float, float]):
        object_values: List[Tuple[int, Tuple[float, float, float, float]]] = self.values[object_id]
        index: int = bisect_left(object_values, position, key=lambda x: x[0])
        if index < len(object_values) and object_values[index][0] == position:
            # There is already an area for this position; replace it.
            object_values[index] = (position, area)
        else:
            object_values.insert(index, (position, area))

    def add_object(self, index: int, object_id: str, name: str):
        self.objects.append((index, object_id, name))
        self.values[object_id] = []

    def remove_object(self, object_id: str):
        del self.values[object_id]
        for i, (_index, some_object_id, _name) in enumerate(self.objects):
            if object_id == some_object_id:
                del self.objects[i]
                break

    def interpolate(self, object_id: str, position: int) -> Optional[Tuple[float, float, float, float]]:
        object_values: List[Tuple[int, Tuple[float, float, float, float]]] = self.values[object_id]
        if not object_values:
            return None

        index: int = bisect_left(object_values, position, key=lambda x: x[0])
        if index == 0:
            # Return the first area.
            return object_values[0][1]
        elif index < len(object_values):
            # Return the interpolated area.
            xp = (object_values[index - 1][0], object_values[index][0])
            yps = zip(object_values[index - 1][1], object_values[index][1])
            return tuple(numpy.interp(position, xp, yp) for yp in yps)
        else:
            # Return the last area.
            return object_values[-1][1]

    def greatest_index(self) -> int:
        if self.objects:
            return max(index for index, _object_id, _name in self.objects)
        else:
            return 0


class TrackedObjectItem(GObject.GObject):
    """Data for displaying a Tracked Object in the list.

    Attributes:
        object_id (str): Identifier to be passed externally where the data
            might have to be reused.
        name (str): The name for display.
        index (int): Internal identifier for sorting the objects in the
            order they have been created.
    """

    def __init__(self, object_id: str, index: int, name: str):
        GObject.GObject.__init__(self)
        self.object_id: str = object_id
        self.index: int = index
        self.name: str = name


class TrackedObjectRow(Gtk.ListBoxRow):
    """Represents a tracked object to be selected in a list."""

    def __init__(self, object_id: str, name: str):
        Gtk.ListBoxRow.__init__(self)
        self.object_id: str = object_id
        self.name: str = name

        label = Gtk.Label(name)
        label.props.margin = SPACING
        label.props.margin_end = PADDING
        label.props.margin_start = PADDING
        label.props.halign = Gtk.Align.START
        label.show()
        self.add(label)


@Gtk.Template(filename=os.path.join(get_ui_dir(), "trackerperspective.ui"))
class ToplevelWidget(Gtk.Box, Loggable):
    """Toplevel widget of the Tracker perspective."""

    __gtype_name__ = "ToplevelWidget"

    add_object_button = Gtk.Template.Child()
    algorithm_combo_box = Gtk.Template.Child()
    aspect_frame = Gtk.Template.Child()
    drawing_area = Gtk.Template.Child()
    howto_add_infobar = Gtk.Template.Child()
    next_frame_button = Gtk.Template.Child()
    object_listbox = Gtk.Template.Child()
    object_manager_box = Gtk.Template.Child()
    pause_icon = Gtk.Template.Child()
    play_icon = Gtk.Template.Child()
    play_pause_button = Gtk.Template.Child()
    pos_adj = Gtk.Template.Child()
    prev_frame_button = Gtk.Template.Child()
    remove_object_button = Gtk.Template.Child()
    seeker = Gtk.Template.Child()
    stop_button = Gtk.Template.Child()
    track_button = Gtk.Template.Child()
    viewer_buttons = Gtk.Template.Child()
    viewer_overlay = Gtk.Template.Child()

    def __init__(self, app, asset: GES.Asset):
        Gtk.Box.__init__(self)
        Loggable.__init__(self)

        self.app = app
        self.asset: GES.Asset = asset
        self.object_manager: ObjectManager = ObjectManager(self.asset)

        info = asset.get_info()
        video_streams = info.get_video_streams()
        stream = video_streams[0]
        self.source_width = stream.get_natural_width()
        self.source_height = stream.get_natural_height()
        self.videorate = self.app.project_manager.current_project.videorate

        self.pipeline = AssetPipeline(self.asset.props.id)
        self.pipeline.connect("error", self._pipeline_error_cb)
        self.pipeline.activate_position_listener(50)
        self.pipeline.connect("position", self._pipeline_position_cb)
        self.pipeline.connect("eos", self._pipeline_eos_cb)
        self.pipeline.connect("state-change", self._pipeline_state_change_cb)
        self.step = (self.videorate.denom / self.videorate.num) * Gst.SECOND

        # The area selected by the user with drag&drop.
        # The coordinates are in screen pixels.
        self.x1: Optional[float] = None  # pylint: disable=invalid-name
        self.y1: Optional[float] = None  # pylint: disable=invalid-name
        self.x2: Optional[float] = None  # pylint: disable=invalid-name
        self.y2: Optional[float] = None  # pylint: disable=invalid-name
        # The size of the viewer when the user selected the area.
        self.drawing_area_width: Optional[int] = None
        self.drawing_area_height: Optional[int] = None

        self.current_object: Optional[str] = None

        self.sink_widget = None
        self.tracker_pipeline: Optional[SimplePipeline] = None
        self.tracker_sink_widget = None
        # Data gathered during the current tracking operation.
        self.roi_data: Optional[Dict[int, Tuple[float, float, float, float]]] = None
        # Position where the last tracking has been started.
        self.start_pos: Optional[int] = None

        # Setup ListBox with the tracked objects
        self.tracked_objects_store = Gio.ListStore()
        for index, object_id, name in self.object_manager.objects:
            self.tracked_objects_store.append(TrackedObjectItem(object_id, index, name))
        self.object_listbox.bind_model(self.tracked_objects_store, self.create_tracked_object_row_func)

        fix_infobar(self.howto_add_infobar)

        # Setup Viewer
        _, self.sink_widget = self.pipeline.create_sink()
        self.aspect_frame.set(
            xalign=0.5, yalign=0.5, ratio=self.source_width / self.source_height, obey_child=False)
        self.viewer_overlay.add(self.sink_widget)
        self.viewer_overlay.add_overlay(self.drawing_area)
        self.viewer_overlay.show_all()

        # Setup Seeker
        self.seeker.props.adjustment.set_upper(self.asset.props.duration)
        self.seeker.props.adjustment.set_step_increment(self.step)

        # Setup algorithm ComboBox
        cell = Gtk.CellRendererText()
        self.algorithm_combo_box.set_model(self.__get_tracking_algorithms())
        self.algorithm_combo_box.pack_start(cell, False)
        self.algorithm_combo_box.add_attribute(cell, "text", 0)

        self._setup(None)

    @Gtk.Template.Callback()
    def _viewer_overlay_realize_cb(self, widget):
        self.pipeline.pause()

    # Playback methods

    def _pipeline_error_cb(self, pipeline, message, detail):
        self.warning("pipeline error: %s (%s)", message, detail)
        self.pipeline.set_simple_state(Gst.State.NULL)

    def __update_adjustment(self, position: int):
        """Updates the UI without triggering callbacks."""
        self.pos_adj.handler_block_by_func(self._adjustment_value_changed_cb)
        try:
            self.pos_adj.set_value(position)
        finally:
            self.pos_adj.handler_unblock_by_func(self._adjustment_value_changed_cb)

    def _pipeline_position_cb(self, pipline, position):
        self.__update_adjustment(position)

    def _pipeline_eos_cb(self, pipeline):
        pipeline.simple_seek(0)

    def _pipeline_state_change_cb(self, pipeline, state, prev_state):
        self.log("Pipeline state %s -> %s", prev_state, state)
        if pipeline.playing():
            icon = self.pause_icon
        else:
            icon = self.play_icon

        self.play_pause_button.set_image(icon)
        self.track_button.props.sensitive = False

    @Gtk.Template.Callback()
    def _play_pause_button_clicked_cb(self, button):
        self.pipeline.toggle_playback()
        self.__reset_selected_area()

    @Gtk.Template.Callback()
    def _next_frame_button_clicked_cb(self, button):
        self._seek(1)

    @Gtk.Template.Callback()
    def _prev_frame_button_clicked_cb(self, button):
        self._seek(-1)

    def _seek(self, direction: int):
        state = self.pipeline.get_simple_state()
        if state == Gst.State.PLAYING:
            self.pipeline.pause()
        elif state == Gst.State.PAUSED:
            self.pipeline.seek_relative(self.step * direction)

    @Gtk.Template.Callback()
    def _adjustment_value_changed_cb(self, adjustment):
        """Handle a seek performed by the user interacting with the UI."""
        if self.pipeline.get_simple_state() != Gst.State.PAUSED:
            self.pipeline.pause()

        # Block the pipeline's "position" signal to prevent a callback loop.
        self.pipeline.handler_block_by_func(self._pipeline_position_cb)
        try:
            self.pipeline.simple_seek(adjustment.props.value)
        finally:
            self.pipeline.handler_unblock_by_func(self._pipeline_position_cb)

    # Bounding box callbacks

    @Gtk.Template.Callback()
    def _drawing_area_button_event_cb(self, widget, event):
        res, button = event.get_button()
        if not res or button != 1:
            return

        if self.pipeline.get_simple_state() != Gst.State.PAUSED:
            return

        if event.get_event_type() == Gdk.EventType.BUTTON_PRESS:
            self.x1 = event.x
            self.y1 = event.y
            self.drawing_area_width = self.drawing_area.get_allocated_width
            self.drawing_area_height = self.drawing_area.get_allocated_height
        elif event.get_event_type() == Gdk.EventType.BUTTON_RELEASE:
            self.x2 = event.x
            self.y2 = event.y

            # Convert the viewer coordinates to video coordinates.
            factor: float = 1 / self.video_to_viewer_factor()
            x = min(self.x1, self.x2) * factor
            y = min(self.y1, self.y2) * factor
            w = abs(self.x2 - self.x1) * factor
            h = abs(self.y2 - self.y1) * factor
            if w and h:
                # Apply the selected area to the object.
                if not self.current_object:
                    self._create_empty_object()
                position = self.pipeline.get_position()
                self.object_manager.update_object_position(
                    self.current_object, position, (x, y, w, h))

                # Allow tracking.
                self.track_button.props.sensitive = True

            self.__reset_selected_area()

    def _create_empty_object(self):
        """Generates a new object and adds it to the object_manager."""
        index = 1 + self.object_manager.greatest_index()
        object_id = uuid.uuid4().hex
        name = _("Object {}").format(index)

        self.current_object = object_id
        self.object_manager.add_object(index, object_id, name)

        self.tracked_objects_store.append(TrackedObjectItem(object_id, index, name))
        last_row_index = self.tracked_objects_store.get_n_items() - 1
        last_row = self.object_listbox.get_row_at_index(last_row_index)
        self.object_listbox.select_row(last_row)

    def __reset_selected_area(self):
        self.x1, self.y1 = None, None
        self.x2, self.y2 = None, None
        self.drawing_area_width, self.drawing_area_height = None, None

    @Gtk.Template.Callback()
    def _drawing_area_enter_notify_event_cb(self, widget, event):
        self.app.gui.get_window().set_cursor(Gdk.Cursor(Gdk.CursorType.CROSSHAIR))

    @Gtk.Template.Callback()
    def _drawing_area_leave_notify_event_cb(self, widget, event):
        self.app.gui.get_window().set_cursor(NORMAL_CURSOR)

    @Gtk.Template.Callback()
    def _drawing_area_motion_notify_event_cb(self, widget, event):
        if self.x1 is not None:
            self.x2 = event.x
            self.y2 = event.y
            self.drawing_area.queue_draw()

    @Gtk.Template.Callback()
    def _drawing_area_draw_cb(self, drawing_area, cr):
        """Handler responsible for drawing the selection rectangle."""
        if self.x2 is not None:
            # Draw the area being delimited by the user on the viewer.
            x, y = self.x1, self.y1
            w, h = self.x2 - self.x1, self.y2 - self.y1
        elif self.current_object:
            # Draw the area tracked previously for the current position.
            position = self.pipeline.get_position(fails=False)
            video_coords = self.object_manager.interpolate(self.current_object, position)
            if not video_coords:
                return

            x, y, w, h = video_coords

            # Translate from video coordinates to viewer coordinates.
            factor: float = self.video_to_viewer_factor()
            cr.scale(factor, factor)
        else:
            # Nothing to draw.
            return

        cr.set_operator(cairo.OPERATOR_OVER)

        cr.set_source_rgba(1, 1, 0.6, 0.8)  # Yellow
        cr.rectangle(x, y, w, h)
        cr.stroke()

        cr.set_source_rgba(0.6, 0.6, 0.6, 0.5)  # Light gray
        cr.rectangle(x, y, w, h)
        cr.fill()

    def video_to_viewer_factor(self) -> float:
        if self.source_width > self.source_height:
            viewer_width = self.drawing_area.get_allocated_width()
            return viewer_width / self.source_width
        else:
            viewer_height = self.drawing_area.get_allocated_height()
            return viewer_height / self.source_height

    @Gtk.Template.Callback()
    def _add_object_button_clicked_cb(self, button):
        # If no object is selected then the user will be able to
        # delimit an area to create an object.
        self.object_listbox.unselect_all()

    @Gtk.Template.Callback()
    def _remove_object_button_clicked_cb(self, button):
        row = self.object_listbox.get_selected_row()
        index = row.get_index()
        tracked_object: TrackedObjectItem = self.tracked_objects_store.get_item(index)
        self.tracked_objects_store.remove(index)

        self.remove_object_button.props.sensitive = False
        self.current_object = None

        self.object_manager.remove_object(tracked_object.object_id)
        self.seeker.clear_marks()

    @Gtk.Template.Callback()
    def _stop_track_button_clicked_cb(self, button):
        self.tracker_pipeline.pause()
        self.__stop_tracker()

    def _setup(self, object_id: Optional[str]):
        """Sets up the UI for creating or updating a tracked object."""
        self.seeker.clear_marks()

        self._setup_tracking_ui(started=False)

        self.__reset_selected_area()

        self.current_object = object_id
        has_object = bool(object_id)
        if has_object:
            timed_data = self.object_manager.values[object_id]
            if timed_data:
                seek_pos, _area = timed_data[0]
                self.seeker.add_mark(seek_pos, Gtk.PositionType.BOTTOM, None)
                self.pipeline.simple_seek(seek_pos)

        self.add_object_button.props.sensitive = has_object
        self.remove_object_button.props.sensitive = has_object
        # Hide the infobar by making it transparent. This way the left column
        # of widgets has a stable width, as the infobar is the widest widget.
        self.howto_add_infobar.props.opacity = 1 if not has_object else 0

    # Object list box methods

    def create_tracked_object_row_func(self, item: TrackedObjectItem) -> TrackedObjectRow:
        return TrackedObjectRow(item.object_id, item.name)

    @Gtk.Template.Callback()
    def _listbox_selected_rows_changed_cb(self, listbox: Gtk.ListBox):
        row: TrackedObjectRow = listbox.get_selected_row()
        self._setup(row.object_id if row else None)

    # Tracker methods

    @Gtk.Template.Callback()
    def _track_button_clicked_cb(self, button):
        self._setup_tracking_ui(started=True)

        # Build the object tracking pipeline.
        algorithm = self.algorithm_combo_box.get_active()
        self.start_pos = self.pipeline.get_position()
        x, y, w, h = self.object_manager.interpolate(self.current_object, self.start_pos)
        self.roi_data = {}
        _pipeline = Gst.parse_launch(
            "uridecodebin uri={} ! videoconvert ! \
            cvtracker object-initial-x={} object-initial-y={} object-initial-width={} \
            object-initial-height={} algorithm={} draw-rect=true ! tee name=t ! \
            queue ! videoconvert ! gtksink name=gtksink t. \
            ! fakesink name=sink signal-handoffs=TRUE"
            .format(self.asset.props.id, int(x), int(y), int(w), int(h), algorithm))

        self.seeker.add_mark(self.start_pos, Gtk.PositionType.BOTTOM, None)

        # Connect to fakesink to get the tracking data.
        fakesink = _pipeline.get_by_name("sink")
        fakesink.connect("handoff", self.__tracker_handoff_cb, self.roi_data)

        # Set up a widget to show the video as the object is being tracked.
        video_sink = _pipeline.get_by_name("gtksink")
        self.tracker_sink_widget = video_sink.props.widget
        self.viewer_overlay.add_overlay(self.tracker_sink_widget)

        # Create a high-level pipeline to get position updates.
        self.tracker_pipeline = SimplePipeline(_pipeline)
        self.tracker_pipeline.activate_position_listener(50)
        self.tracker_pipeline.connect("position", self.__tracker_position_cb)

        # Connect to the bus of the pipeline to find out when the stream ends.
        bus = _pipeline.get_bus()
        bus.connect("message", self.__tracker_bus_message_cb)
        bus.add_signal_watch()

        # Start the tracking pipeline.
        self.tracker_pipeline.simple_seek(self.start_pos)
        self.tracker_pipeline.play()

    def _setup_tracking_ui(self, started: bool):
        """Sets up the widgets depending on the specified tracking status."""
        self.drawing_area.props.visible = not started
        self.object_manager_box.props.sensitive = not started
        self.algorithm_combo_box.props.sensitive = not started
        self.viewer_buttons.props.sensitive = not started

        self.track_button.props.visible = not started
        self.track_button.props.sensitive = False
        self.stop_button.props.visible = started

    def __get_tracking_algorithms(self) -> Gtk.ListStore:
        listmodel = Gtk.ListStore(str)
        element = Gst.ElementFactory.make("cvtracker", "tracker")
        properties = element.list_properties()
        for prop in properties:
            if prop.name == "algorithm":
                for unused_key, algorithm in prop.enum_class.__enum_values__.items():
                    listmodel.append([algorithm.value_nick])
                break

        return listmodel

    def __stop_tracker(self):
        position = self.tracker_pipeline.get_position(fails=False)
        self.pipeline.simple_seek(position)

        self.log("Waiting for the tracker_pipeline to stop")
        self.tracker_pipeline.connect("state-change", self.__pipeline_state_change_cb)
        self.tracker_pipeline.stop()

        self.object_manager.update_object(self.current_object, self.start_pos, self.roi_data)
        self.start_pos = None
        self.roi_data = None

    def __pipeline_state_change_cb(self, pipeline, state, prev_state):
        if state != Gst.State.READY:
            return

        self.log("Tracker_pipeline stopped")
        self.tracker_pipeline.disconnect_by_func(self.__pipeline_state_change_cb)
        self.tracker_pipeline.release()
        self.tracker_pipeline = None

        self.viewer_overlay.remove(self.tracker_sink_widget)

        self._setup_tracking_ui(started=False)

    def __tracker_position_cb(self, pipeline, position):
        if position >= self.start_pos:
            if not self.tracker_sink_widget.props.visible:
                self.tracker_sink_widget.show()

            self.__update_adjustment(position)

    def __tracker_bus_message_cb(self, bus, message):
        if message.type == Gst.MessageType.EOS:
            self.__stop_tracker()

    def __tracker_handoff_cb(self, element, buffer, pad, roi_data: Dict[int, Tuple[float, float, float, float]]):
        video_roi = GstVideo.buffer_get_video_region_of_interest_meta_id(buffer, 0)
        if video_roi:
            roi_data[buffer.pts] = (video_roi.x, video_roi.y, video_roi.w, video_roi.h)
        else:
            self.log("lost tracker at: %s", buffer.pts / Gst.SECOND)


class TrackerPerspective(Perspective):
    """Pitivi's Tracker Perspective.

    Allows the user to track multiple objects
    and manually correct the obtained track data.

    Attributes:
        app (Pitivi): The app.
        asset (GES.UriClipAsset): Asset to be used.
    """

    def __init__(self, app, asset):
        super().__init__()
        self.app = app
        self.asset = asset

    def __create_headerbar(self):
        headerbar = Gtk.HeaderBar()
        headerbar.set_show_close_button(False)

        back_button = Gtk.Button.new_from_icon_name(
            "go-previous-symbolic", Gtk.IconSize.SMALL_TOOLBAR)
        back_button.set_always_show_image(True)
        back_button.set_tooltip_text(_("Go back"))
        back_button.connect("clicked", self.__back_button_clicked_cb)
        back_button.set_margin_end(4 * PADDING)
        headerbar.pack_start(back_button)
        headerbar.props.title = os.path.basename(self.asset.props.id)
        headerbar.show_all()

        return headerbar

    def setup_ui(self):
        self.toplevel_widget = ToplevelWidget(self.app, self.asset)
        self.headerbar = self.__create_headerbar()

    def __back_button_clicked_cb(self, button):
        self.toplevel_widget.object_manager.save()
        self.toplevel_widget.pipeline.release()
        self.app.gui.show_perspective(self.app.gui.editor)

    def refresh(self):
        """Refreshes the perspective."""
        self.toplevel_widget.play_pause_button.grab_focus()


class CoverObjectPopover(Gtk.Popover, Loggable):
    """Popover for selecting an object to cover."""

    # The representation of the effect providing the cover.
    _EFFECT_PIPELINE = "video videotestsrc pattern=solid-color foreground-color=0xff000000 ! framepositioner name=positioner ! gescompositor"

    def __init__(self, app, clip: GES.Clip):
        Gtk.Popover.__init__(self)
        Loggable.__init__(self)

        self.app = app

        self.clip: GES.Clip = clip
        self.object_manager: Optional[ObjectManager] = None

        self.listbox = Gtk.ListBox()
        self.listbox.connect("row-activated", self.__row_activated_cb)

        self.scroll_window = Gtk.ScrolledWindow()
        self.scroll_window.add(self.listbox)
        self.scroll_window.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self.scroll_window.props.max_content_height = 350
        self.scroll_window.props.propagate_natural_height = True

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, margin=PADDING)
        vbox.pack_start(self.scroll_window, True, True, 0)
        vbox.show_all()

        self.add(vbox)

    def update_object_list(self):
        """Updates the list of not yet covered objects."""
        self.object_manager = ObjectManager(self.clip.asset)

        for row in self.listbox.get_children():
            self.listbox.remove(row)

        # Check which tracked objects have already been covered.
        covered_objects = []
        for effect in self.clip.get_top_effects():
            tracked_object_id = effect.get_string(EFFECT_TRACKED_OBJECT_ID_META)
            if tracked_object_id:
                covered_objects.append(tracked_object_id)

        # Allow selecting the not-yet-covered objects.
        for _index, object_id, name in self.object_manager.objects:
            if object_id not in covered_objects:
                self.listbox.add(TrackedObjectRow(object_id, name))

        # Allow tracking new objects.
        button_row = Gtk.ListBoxRow(selectable=False)
        track_objects_button = Gtk.Button(_("Track objects"))
        track_objects_button.connect("clicked", self.__track_objects_button_clicked_cb)
        button_row.add(track_objects_button)
        self.listbox.add(button_row)

        self.listbox.show_all()

    def __row_activated_cb(self, listbox: Gtk.ListBox, row: TrackedObjectRow):
        self._create_effect(row.object_id, row.name)

        self.popdown()

    def __effect_control_binding_added_cb(self, track_element, binding, object_id):
        control_source = binding.props.control_source
        timed_data = self.object_manager.values[object_id]
        for timestamp, (x, y, w, h) in timed_data:
            if binding.name == "posx":
                value = x
            elif binding.name == "posy":
                value = y
            elif binding.name == "width":
                value = w
            elif binding.name == "height":
                value = h
            else:
                break

            control_source.set(timestamp, value)

    def __clip_child_added_cb(self, clip, track_element, object_id):
        if not isinstance(track_element, GES.Effect):
            return

        clip.disconnect_by_func(self.__clip_child_added_cb)

        track_element.connect("control-binding-added",
                              self.__effect_control_binding_added_cb,
                              object_id)
        try:
            for prop in ("posx", "posy", "width", "height"):
                control_source = GstController.InterpolationControlSource()
                control_source.props.mode = GstController.InterpolationMode.NONE
                track_element.set_control_source(control_source, prop, "direct-absolute")
        finally:
            track_element.disconnect_by_func(self.__effect_control_binding_added_cb)

    def _create_effect(self, object_id: str, name: str):
        effect = GES.Effect.new(self._EFFECT_PIPELINE)
        effect.register_meta_string(GES.MetaFlag.READABLE, EFFECT_TRACKED_OBJECT_ID_META, object_id)
        effect.register_meta_string(GES.MetaFlag.READABLE, EFFECT_TRACKED_OBJECT_NAME_META, name)

        self.log("Waiting for effect to be added to the clip")
        self.clip.connect("child-added", self.__clip_child_added_cb, object_id)
        self.clip.add_top_effect(effect, 0)

        self.app.project_manager.current_project.pipeline.commit_timeline()

    def __track_objects_button_clicked_cb(self, button):
        tracker = TrackerPerspective(self.app, self.clip.asset)
        self.app.project_manager.current_project.pipeline.pause()
        tracker.setup_ui()
        self.app.gui.show_perspective(tracker)



# -*- coding: utf-8 -*-
# Copyright (c) 2017, Fabian Orccon <cfoch.fabian@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""PluginManager for loading Pitivi plugins."""
import os
from enum import IntEnum
from gettext import gettext as _

from gi.repository import GObject
from gi.repository import Peas

from pitivi.configure import get_plugins_dir
from pitivi.configure import get_user_plugins_dir
from pitivi.settings import GlobalSettings
from pitivi.utils.loggable import Loggable


GlobalSettings.add_config_section("plugins")
GlobalSettings.add_config_option("ActivePlugins",
                                 section="plugins", key="active-plugins",
                                 default=[])


class API(GObject.GObject):
    """Interface that gives access to all the objects inside Pitivi."""

    def __init__(self, app):
        GObject.GObject.__init__(self)
        self.app = app


class PluginType(IntEnum):
    """Types of plugins we support, depending on their directory location."""

    # pylint: disable=comparison-with-callable,inconsistent-return-statements,no-else-return

    SYSTEM = 1
    USER = 2

    def __str__(self):
        if self.value == PluginType.USER:
            return _("User plugins")
        elif self.value == PluginType.SYSTEM:
            return _("System plugins")

    def get_dir(self):
        """Returns the directory where this type of plugins can be found."""
        if self.value == PluginType.USER:
            return get_user_plugins_dir()
        elif self.value == PluginType.SYSTEM:
            return get_plugins_dir()


class PluginManager(Loggable):
    """Pitivi Plugin Manager to handle a set of plugins.

    Attributes:
        DEFAULT_LOADERS (tuple): Default loaders used by the plugin manager. For
            possible values see
            https://developer.gnome.org/libpeas/stable/PeasEngine.html#peas-engine-enable-loader
    """

    DEFAULT_LOADERS = ("python3", )

    def __init__(self, app):
        Loggable.__init__(self)
        self.app = app
        self.engine = Peas.Engine.get_default()
        # Many plugins need access to the main window. However, by the time a
        # plugin is loaded from settings (as soon Pitivi application starts),
        # the main window doesn't exist yet. So we load plugins from settings
        # after the main window is added.
        self.app.connect("window-added", self.__window_added_cb)

        for loader in self.DEFAULT_LOADERS:
            self.engine.enable_loader(loader)

        self._setup_plugins_dir()
        self._setup_extension_set()

    @property
    def plugins(self):
        """Gets the engine's plugin list."""
        return self.engine.get_plugin_list()

    @classmethod
    def get_plugin_type(cls, plugin_info):
        """Gets the PluginType for the specified Peas.PluginInfo."""
        paths = [plugin_info.get_data_dir(), get_plugins_dir()]
        if os.path.commonprefix(paths) == get_plugins_dir():
            return PluginType.SYSTEM
        return PluginType.USER

    def get_extension(self, module_name):
        """Gets the extension identified by the specified name.

        Args:
            module_name (str): The name of the extension.

        Returns:
            The extension if exists. Otherwise, `None`.
        """
        plugin = self.get_plugin_info(module_name)
        if not plugin:
            return None

        return self.extension_set.get_extension(plugin)

    def get_plugin_info(self, module_name):
        """Gets the plugin info for the specified plugin name.

        Args:
            module_name (str): The name from the .plugin file of the module.

        Returns:
            Peas.PluginInfo: The plugin info if it exists. Otherwise, `None`.
        """
        for plugin in self.plugins:
            if plugin.get_module_name() == module_name:
                return plugin
        return None

    def _load_plugins(self):
        """Loads plugins from settings."""
        plugin_names = self.app.settings.ActivePlugins
        for plugin_name in plugin_names:
            plugin_info = self.engine.get_plugin_info(plugin_name)
            if plugin_info not in self.plugins:
                self.warning("Plugin missing: %s", plugin_name)
                continue
            self.engine.load_plugin(plugin_info)

    def _setup_extension_set(self):
        plugin_iface = API(self.app)
        self.extension_set = Peas.ExtensionSet.new(self.engine,
                                                   Peas.Activatable,
                                                   ["object"],
                                                   [plugin_iface])
        self.extension_set.connect("extension-removed",
                                   self.__extension_removed_cb)
        self.extension_set.connect("extension-added",
                                   self.__extension_added_cb)

    def _setup_plugins_dir(self):
        plugins_dir = get_plugins_dir()
        user_plugins_dir = get_user_plugins_dir()
        if os.path.exists(plugins_dir):
            self.engine.add_search_path(plugins_dir)
        if os.path.exists(user_plugins_dir):
            self.engine.add_search_path(user_plugins_dir)

    @staticmethod
    def __extension_removed_cb(unused_set, unused_plugin_info, extension):
        extension.deactivate()

    @staticmethod
    def __extension_added_cb(unused_set, unused_plugin_info, extension):
        extension.activate()

    def __window_added_cb(self, unused_app, unused_window):
        """Handles the addition of a window to the application."""
        self._load_plugins()
        self.engine.connect("notify::loaded-plugins", self.__loaded_plugins_cb)
        self.app.disconnect_by_func(self.__window_added_cb)

    def __loaded_plugins_cb(self, engine, unused_pspec):
        """Handles the changing of the loaded plugin list."""
        self.app.settings.ActivePlugins = engine.get_property("loaded-plugins")



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2014, Thibault Saunier <thibault.saunier@collabora.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
# pylint: disable=protected-access
import locale
import subprocess
from unittest import mock

import gi
from gi.repository import Gdk
from gi.repository import GES
from gi.repository import GLib
from gi.repository import Gst
from gi.repository import Gtk

from pitivi.utils.timeline import Zoomable


CAT = "validate"


try:
    gi.require_version("GstValidate", "1.0")
    from gi.repository import GstValidate
except ImportError:
    GstValidate = None
except ValueError:
    # pylint: disable=invalid-name
    GstValidate = None

monitor = None
has_validate = False


def create_event(event_type, **kwargs):
    event_types_constructors = {
        Gdk.EventType.BUTTON_PRESS: Gdk.EventButton,
        Gdk.EventType.BUTTON_RELEASE: Gdk.EventButton,
        Gdk.EventType.KEY_PRESS: Gdk.EventKey,
        Gdk.EventType.KEY_RELEASE: Gdk.EventKey,
        Gdk.EventType.MOTION_NOTIFY: Gdk.EventMotion
    }

    constructor = event_types_constructors[event_type]
    event = constructor()
    event.type = event_type
    for arg, value in kwargs.items():
        setattr(event, arg, value)

    return event


if GstValidate:
    class PitiviMonitor(GstValidate.Monitor):
        def __init__(self, runner, obj):
            GstValidate.Monitor.__init__(self, object=obj, validate_runner=runner)

            if GstValidate:
                try:
                    gi.require_version('Wnck', '3.0')
                    from gi.repository import Wnck
                    Wnck.Screen.get_default().connect("window-opened", self._window_opened_cb)
                except (ImportError, ValueError):
                    print("Wnck not present on the system,"
                          " not checking the sink does not open a new window")
                except AttributeError:
                    print("Wnck can not be used on the system")

        def _window_opened_cb(self, screen, window):
            global monitor

            if window.get_name() == 'renderer' and monitor:
                monitor.report_simple(GLib.quark_from_string("pitivi::wrong-window-creation"),
                                      "New window created by the sink,"
                                      " that should not happen")

        def check_wrong_window(self):
            try:
                windows = subprocess.check_output(["xwininfo", "-tree", "-root"]).decode(locale.getdefaultlocale()[1])
                for w in windows.split('\n'):
                    if "OpenGL renderer" in w and w.startswith("     0x"):
                        monitor.report_simple(GLib.quark_from_string("pitivi::wrong-window-creation"),
                                              "New window created by the sink,"
                                              " that should not happen, (current windows: %s)"
                                              % windows)
                        break
            except (subprocess.CalledProcessError, FileNotFoundError):
                pass


def create_monitor(runner, app):
    global monitor
    global has_validate

    if not monitor and has_validate:
        monitor = PitiviMonitor(runner, app)
        GstValidate.Reporter.set_name(monitor, "Pitivi")


def stop_func(scenario, action):
    global monitor

    if monitor:
        monitor.check_wrong_window()

    if action.structure.get_boolean("force")[0]:
        GstValidate.execute_action(GstValidate.get_action_type(action.type).overriden_type,
                                   action)

        timeline = scenario.get_pipeline().props.timeline
        project = timeline.get_asset()

        if project:
            project.set_modification_state(False)
            GstValidate.print_action(action, "Force quitting, ignoring any"

                                     " changes in the project\n")
        timeline.ui.app.shutdown()

        return 1

    GstValidate.print_action(action, "STOP: not doing anything in pitivi\n")

    return 1


def position_changed_cb(pipeline, position, scenario, action,
                        wanted_position):
    if pipeline._busy_async:
        return

    if pipeline._next_seek:
        return

    print(str(wanted_position), str(position))
    if wanted_position != position:
        scenario.report_simple(
            GLib.quark_from_string("scenario::execution-error"),
            "Position after seek (%s) does not match wanted one %s" % (
                Gst.TIME_ARGS(position), Gst.TIME_ARGS(wanted_position)))

    pipeline.disconnect_by_func(position_changed_cb)
    action.set_done()


def seek_func(scenario, action):
    res, wanted_position = GstValidate.utils_get_clocktime(action.structure,
                                                           "start")
    assert res
    scenario.get_pipeline().simple_seek(wanted_position)
    scenario.get_pipeline().connect("position", position_changed_cb, scenario,
                                    action, wanted_position)

    return GstValidate.ActionReturn.ASYNC


def set_state_func(scenario, action):
    wanted_state = action.structure["state"]
    if wanted_state is None:
        wanted_state = action.structure.get_name()
        if wanted_state == "play":
            wanted_state = "playing"
        elif wanted_state == "pause":
            wanted_state = "paused"

    if wanted_state == "paused":
        if scenario.__dict__.get("started", None) is None:
            return 1

    return GstValidate.execute_action(GstValidate.get_action_type(action.type).overriden_type,
                                      action)


def get_edge(structure):
    try:
        res, edge = GstValidate.utils_enum_from_str(GES.Edge, structure["edge"])
        if not res:
            edge = GES.Edge.EDGE_NONE
        else:
            edge = GES.Edge(edge)

    except KeyError:
        edge = GES.Edge.EDGE_NONE

    return edge


def _release_button_if_needed(scenario, timeline, container, layer_prio,
                              position, y):
    next_actions = scenario.get_actions()
    if len(next_actions) == 1:
        return
    next_action = None
    for next_action in next_actions[1:]:
        if next_action.type not in ["wait", "add-layer"]:
            break
    if not next_action:
        return

    need_release = True
    if next_action and next_action.type == "edit-container":
        edge = get_edge(next_action.structure)

        if edge == scenario.last_edge:
            need_release = False

    if next_action is None or need_release:
        scenario.dragging = False
        x = Zoomable.ns_to_pixel_accurate(position)
        event = create_event(Gdk.EventType.BUTTON_RELEASE, button=1, x=x, y=y)
        with mock.patch.object(Gtk, "get_event_widget") as get_event_widget:
            get_event_widget.return_value = container.ui
            container.ui._button_release_event_cb(None, event)

        if layer_prio > 0 and container.get_layer().get_priority() != layer_prio:
            scenario.report_simple(GLib.quark_from_string("scenario::execution-error"),
                                   "Resulting clip priority: %s"
                                   " is not the same as the wanted one: %s"
                                   % (container.get_layer().get_priority(),
                                      layer_prio))

        clean_edit_modes(timeline, scenario)


def clean_edit_modes(timeline, scenario):
    if scenario.last_mode == GES.EditMode.EDIT_RIPPLE:
        event = create_event(Gdk.EventType.KEY_RELEASE, keyval=Gdk.KEY_Shift_L)
        timeline.ui.get_parent().do_key_release_event(event)
    elif scenario.last_mode == GES.EditMode.EDIT_ROLL:
        event = create_event(Gdk.EventType.KEY_RELEASE, keyval=Gdk.KEY_Control_L)
        timeline.ui.get_parent().do_key_release_event(event)

    scenario.last_mode = None


def set_editing_mode(timeline, scenario, action):
    if not hasattr(scenario, "last_mode"):
        scenario.last_mode = None

    try:
        res, mode = GstValidate.utils_enum_from_str(GES.EditMode, action.structure["edit-mode"])
        if not res:
            mode = GES.EditMode.EDIT_NORMAL
        else:
            mode = GES.EditMode(mode)
    except KeyError:
        mode = GES.EditMode.EDIT_NORMAL

    if mode == GES.EditMode.EDIT_RIPPLE:
        timeline.ui.get_parent().do_key_press_event(create_event(Gdk.EventType.KEY_PRESS, keyval=Gdk.KEY_Shift_L))

        if scenario.last_mode == GES.EditMode.EDIT_ROLL:
            timeline.ui.get_parent().do_key_release_event(create_event(Gdk.EventType.KEY_RELEASE, keyval=Gdk.KEY_Control_L))

    elif mode == GES.EditMode.EDIT_ROLL:
        timeline.ui.do_key_press_event(create_event(Gdk.EventType.KEY_PRESS, keyval=Gdk.KEY_Control_L))

        if scenario.last_mode == GES.EditMode.EDIT_RIPPLE:
            timeline.ui.do_key_release_event(create_event(Gdk.EventType.KEY_RELEASE, keyval=Gdk.KEY_Shift_L))
    else:
        clean_edit_modes(timeline, scenario)

    scenario.last_mode = mode


def edit_container_func(scenario, action):
    timeline = scenario.get_pipeline().props.timeline
    container = timeline.get_element(action.structure["container-name"])

    if container is None:
        for layer in timeline.get_layers():
            for clip in layer.get_clips():
                Gst.info("Existing clip: %s" % clip.get_name())

        scenario.report_simple(GLib.quark_from_string("scenario::execution-error"),
                               "Could not find container: %s"
                               % action.structure["container-name"])

        return 1

    res, position = GstValidate.action_get_clocktime(scenario, action, "position")
    layer_prio = action.structure["new-layer-priority"]

    if res is False:
        return 0

    edge = get_edge(action.structure)
    container_ui = container.ui

    set_editing_mode(timeline, scenario, action)

    y = 21 - container_ui.translate_coordinates(timeline.ui, 0, 0)[1]

    if container.get_layer().get_priority() != layer_prio and layer_prio != -1:
        try:
            layer = timeline.get_layers()[layer_prio]
            Gst.info("Y is: %s Realized?? %s Priori: %s layer prio: %s"
                     % (layer.ui.get_allocation().y,
                        container_ui.get_realized(),
                        container.get_layer().get_priority(),
                        layer_prio))
            y = layer.ui.get_allocation().y - container_ui.translate_coordinates(timeline.ui, 0, 0)[1]
            if y < 0:
                y += 21
            elif y > 0:
                y -= 21
        except IndexError:
            if layer_prio == -1:
                y = -5
            else:
                layer = timeline.get_layers()[-1]
                alloc = layer.ui.get_allocation()
                y = alloc.y + alloc.height + 10 - container_ui.translate_coordinates(timeline.ui, 0, 0)[1]

    if not hasattr(scenario, "last_edge"):
        scenario.last_edge = edge

    if not hasattr(scenario, "dragging") or scenario.dragging is False \
            or scenario.last_edge != edge:
        event_widget = container.ui
        if isinstance(container, GES.SourceClip):
            if edge == GES.Edge.EDGE_START:
                event_widget = container.ui.left_handle
            elif edge == GES.Edge.EDGE_END:
                event_widget = container.ui.right_handle

        scenario.dragging = True
        event = create_event(Gdk.EventType.BUTTON_PRESS, button=1, y=y)
        with mock.patch.object(Gtk, "get_event_widget") as get_event_widget:
            get_event_widget.return_value = event_widget
            timeline.ui._button_press_event_cb(event_widget, event)

    x = Zoomable.ns_to_pixel_accurate(position) - container_ui.translate_coordinates(timeline.ui.layout.layers_vbox, 0, 0)[0]
    event = create_event(Gdk.EventType.MOTION_NOTIFY, button=1,
                         x=x, y=y, state=Gdk.ModifierType.BUTTON1_MASK)
    with mock.patch.object(Gtk, "get_event_widget") as get_event_widget:
        get_event_widget.return_value = container.ui
        timeline.ui._motion_notify_event_cb(None, event)

    GstValidate.print_action(action,
                             "Editing %s to %s in %s mode, edge: %s "
                             "with new layer prio: %d\n" % (action.structure["container-name"],
                                                            Gst.TIME_ARGS(position),
                                                            scenario.last_mode,
                                                            edge,
                                                            layer_prio))

    _release_button_if_needed(scenario, timeline, container, layer_prio,
                              position, y)
    scenario.last_edge = edge

    return 1


def split_clip_func(scenario, action):
    timeline = scenario.get_pipeline().props.timeline.ui
    timeline.get_parent()._split_cb(None, None)

    return True


def zoom_func(scenario, action):
    timeline = scenario.get_pipeline().props.timeline

    GstValidate.print_action(action, action.type.replace('-', ' ') + "\n")

    {"zoom-fit": timeline.ui.set_best_zoom_ratio,
     "zoom-out": Zoomable.zoom_out,
     "zoom-in": Zoomable.zoom_in}[action.type]()

    return True


def set_zoom_level_func(scenario, action):
    Zoomable.set_zoom_level(action.structure["level"])

    return True


def add_layer_func(scenario, action):
    timeline = scenario.get_pipeline().props.timeline
    if len(timeline.get_layers()) == 0:
        GstValidate.print_action(action, "Adding first layer\n")
        timeline.append_layer()
    else:
        GstValidate.print_action(action, "Not adding layer, should be done by pitivi itself\n")

    return True


def remove_clip_func(scenario, action):
    try:
        next_action = scenario.get_actions()[1]
    except KeyError:
        next_action = None

    if next_action and next_action.type == "add-clip":
        if next_action.structure["element-name"] == action.structure["element-name"]:
            scenario.no_next_add_element = True
            GstValidate.print_action(action,
                                     "Just moving %s between layers, not removing it\n"
                                     % action.structure["element-name"])
            return True

    action_type = GstValidate.get_action_type(action.type)

    return GstValidate.execute_action(action_type.overriden_type, action)


def select_clips_func(scenario, action):
    should_select = True
    timeline = scenario.get_pipeline().props.timeline
    clip = timeline.get_element(action.structure["clip-name"])

    if clip is None:
        scenario.report_simple(GLib.quark_from_string("scenario::execution-error"),
                               "Could not find container: %s"
                               % action.structure["container-name"])

        return 1

    mode = action.structure["mode"]
    if mode:
        mode = mode.lower()

    if mode == "ctrl":
        if clip.ui.get_state_flags() & Gtk.StateFlags.SELECTED:
            should_select = False

        event = create_event(Gdk.EventType.KEY_PRESS, keyval=Gdk.KEY_Control_L)
        timeline.ui.get_parent().do_key_press_event(event)

    event = create_event(Gdk.EventType.BUTTON_RELEASE, button=1)
    with mock.patch.object(Gtk, "get_event_widget") as get_event_widget:
        get_event_widget.return_value = clip.ui
        clip.ui._button_release_event_cb(None, event)

    selection = action.structure["selection"]
    if not selection:
        if should_select:
            if not clip.ui.get_state_flags() & Gtk.StateFlags.SELECTED:
                scenario.report_simple(GLib.quark_from_string("scenario::execution-error"),
                                       "Clip %s should be selected but is not"
                                       % clip.get_name())
        elif clip.ui.get_state_flags() & Gtk.StateFlags.SELECTED:
            scenario.report_simple(GLib.quark_from_string("scenario::execution-error"),
                                   "Clip %s should be UNselected but is not"
                                   % clip.get_name())
    else:
        for layer in timeline.get_layers():
            for clip in layer.get_clips():
                if clip.get_name() in selection:
                    if not clip.ui.get_state_flags() & Gtk.StateFlags.SELECTED:
                        scenario.report_simple(GLib.quark_from_string("scenario::execution-error"),
                                               "Clip %s should be selected (as defined in selection %s)"
                                               " but is not" % (selection, clip.get_name()))
                else:
                    if clip.ui.get_state_flags() & Gtk.StateFlags.SELECTED:
                        scenario.report_simple(GLib.quark_from_string("scenario::execution-error"),
                                               "Clip %s should NOT be selected (as defined in selection %s)"
                                               " but it is" % (selection, clip.get_name()))

    if mode == "ctrl":
        event = create_event(Gdk.EventType.KEY_RELEASE, keyval=Gdk.KEY_Control_L)
        timeline.ui.get_parent().do_key_release_event(event)

    return 1


def create_action_parameter(name, desc, mandatory=False, possible_variables=None, types=None):
    parameter = GstValidate.ActionParameter()
    parameter.description = desc
    parameter.mandatory = mandatory
    parameter.name = name
    parameter.possible_variables = possible_variables
    parameter.types = types

    return parameter


def init():
    global has_validate
    try:
        GstValidate.init()
        has_validate = GES.validate_register_action_types()
        GstValidate.register_action_type("stop", "pitivi",
                                         stop_func, None,
                                         "Pitivi override for the stop action",
                                         GstValidate.ActionTypeFlags.NONE)

        GstValidate.register_action_type("seek", "pitivi",
                                         seek_func, None,
                                         "Pitivi override for the seek action",
                                         GstValidate.ActionTypeFlags.NONE)

        GstValidate.register_action_type("pause", "pitivi",
                                         set_state_func, None,
                                         "Pitivi override for the pause action",
                                         GstValidate.ActionTypeFlags.NONE)

        GstValidate.register_action_type("play", "pitivi",
                                         set_state_func, None,
                                         "Pitivi override for the pause action",
                                         GstValidate.ActionTypeFlags.NONE)

        GstValidate.register_action_type("set-state", "pitivi",
                                         set_state_func, None,
                                         "Pitivi override for the set-state action",
                                         GstValidate.ActionTypeFlags.NONE)

        GstValidate.register_action_type("edit-container", "pitivi",
                                         edit_container_func, None,
                                         "Start dragging a clip in the timeline",
                                         GstValidate.ActionTypeFlags.NONE)

        GstValidate.register_action_type("split-clip", "pitivi",
                                         split_clip_func, None,
                                         "Split a clip",
                                         GstValidate.ActionTypeFlags.NONE)

        GstValidate.register_action_type("add-layer", "pitivi",
                                         add_layer_func, None,
                                         "Add layer",
                                         GstValidate.ActionTypeFlags.NONE)

        GstValidate.register_action_type("remove-clip", "pitivi",
                                         remove_clip_func, None,
                                         "Remove clip",
                                         GstValidate.ActionTypeFlags.NONE)
        GstValidate.register_action_type("select-clips", "pitivi",
                                         select_clips_func,
                                         [create_action_parameter("clip-name",
                                                                  "The name of the clip to select",
                                                                  True, None, "str")],
                                         "Select clips",
                                         GstValidate.ActionTypeFlags.NONE)

        for action_type in ["zoom-fit", "zoom-out", "zoom-in"]:
            GstValidate.register_action_type(action_type, "pitivi",
                                             zoom_func, None, action_type,
                                             GstValidate.ActionTypeFlags.NO_EXECUTION_NOT_FATAL)
        GstValidate.register_action_type('set-zoom-level', "pitivi", set_zoom_level_func, None, action_type,
                                         GstValidate.ActionTypeFlags.NO_EXECUTION_NOT_FATAL)

        Gst.info("Adding pitivi::wrong-window-creation")
        GstValidate.Issue.register(GstValidate.Issue.new(
            GLib.quark_from_string("pitivi::wrong-window-creation"),
            "A new window for the sink has wrongly been created",
            "All sink should display their images in an embedded "
            "widget and thus not create a new window",
            GstValidate.ReportLevel.CRITICAL))
        return True
    except ImportError:
        has_validate = False
        return False



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2005, Edward Hervey <bilboed@bilboed.com>
# Copyright (c) 2013, 2014, 2015, Thibault Saunier <tsaunier@gnome.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Project related classes."""
import os
import pwd
import shutil
import tarfile
import tempfile
import time
import uuid
from gettext import gettext as _
from hashlib import md5
from typing import Optional
from urllib.parse import unquote

from gi.repository import GdkPixbuf
from gi.repository import GES
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gst
from gi.repository import GstPbutils
from gi.repository import GstVideo
from gi.repository import Gtk

from pitivi.medialibrary import AssetThumbnail
from pitivi.render import Encoders
from pitivi.settings import xdg_cache_home
from pitivi.timeline.previewers import Previewer
from pitivi.timeline.previewers import ThumbnailCache
from pitivi.undo.project import AssetAddedIntention
from pitivi.undo.project import AssetProxiedIntention
from pitivi.utils.loggable import Loggable
from pitivi.utils.misc import disconnect_all_by_func
from pitivi.utils.misc import fixate_caps_with_default_values
from pitivi.utils.misc import is_writable
from pitivi.utils.misc import path_from_uri
from pitivi.utils.misc import quote_uri
from pitivi.utils.misc import scale_pixbuf
from pitivi.utils.misc import unicode_error_dialog
from pitivi.utils.pipeline import Pipeline
from pitivi.utils.ui import beautify_time_delta
from pitivi.utils.ui import SPACING
from pitivi.utils.validate import create_monitor
from pitivi.utils.validate import has_validate


DEFAULT_NAME = _("Untitled")

ALL_RAW_VIDEO_FORMATS = []
# Starting at 2 as 0 is UNKNOWN and 1 is ENCODED.
# We want to make sure we do not try to force ENCODED
# format (as it won't be possible as we have a compositor
# in the pipeline) but we enforce a same VideoFormat is
# used during the whole encoding process.
for i in range(2, GLib.MAXINT):
    try:
        vformat = GstVideo.VideoFormat(i)
        ALL_RAW_VIDEO_FORMATS.append(
            GstVideo.VideoFormat.to_string(vformat))
    except ValueError:
        break

# Properties of encoders that should be ignored when saving/loading
# a project.
IGNORED_PROPS = ["name", "parent"]

# Caps used as the default project settings.
DEFAULT_VIDEO_SETTINGS = "video/x-raw,width=1920,height=1080,framerate=(GstFraction)30/1"
DEFAULT_AUDIO_SETTINGS = "audio/x-raw,channels=2,rate=96000"

# The minimum value for a decent audio rate.
DECENT_AUDIORATE = 44100

# Default values for the safe areas.
DEFAULT_TITLE_AREA_VERTICAL = 0.8
DEFAULT_TITLE_AREA_HORIZONTAL = 0.8
DEFAULT_ACTION_AREA_VERTICAL = 0.9
DEFAULT_ACTION_AREA_HORIZONTAL = 0.9

SCALED_THUMB_WIDTH = 96
SCALED_THUMB_HEIGHT = 54
SCALED_THUMB_DIR = "96x54"
ORIGINAL_THUMB_DIR = "original"


class ProjectManager(GObject.Object, Loggable):
    """The project manager.

    Allows the app to close and then load a different project, handle failures,
    make automatic backups.

    Attributes:
        app (Pitivi): The app.
        current_project (Project): The current project displayed by the app.
        disable_save (bool): Whether save-as is enforced when saving.
    """

    __gsignals__ = {
        "new-project-loading": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "new-project-created": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "new-project-failed": (GObject.SignalFlags.RUN_LAST, None, (str, str)),
        "new-project-loaded": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "save-project-failed": (GObject.SignalFlags.RUN_LAST, None, (str, object)),
        "project-saved": (GObject.SignalFlags.RUN_LAST, None, (object, str)),
        "closing-project": (GObject.SignalFlags.RUN_LAST, bool, (object,)),
        "project-closed": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "missing-uri": (GObject.SignalFlags.RUN_LAST, str, (object, str, object)),
        "reverting-to-saved": (GObject.SignalFlags.RUN_LAST, bool, (object,)),
    }

    def __init__(self, app):
        GObject.Object.__init__(self)
        Loggable.__init__(self)
        self.app = app
        self.current_project = None
        self.disable_save = False
        self._backup_lock = 0
        self.exitcode = 0
        self.__start_loading_time = 0
        self.time_loaded = 0

    def _try_using_backup_file(self, uri):
        backup_path = self._make_backup_uri(path_from_uri(uri))
        use_backup = False
        try:
            path = path_from_uri(uri)
            time_diff = os.path.getmtime(backup_path) - os.path.getmtime(path)
            self.debug(
                'Backup file is %d secs newer: %s', time_diff, backup_path)
        except OSError:
            self.debug('Backup file does not exist: %s', backup_path)
        except UnicodeEncodeError:
            unicode_error_dialog()
        else:
            if time_diff > 0:
                use_backup = self._restore_from_backup_dialog(time_diff)

                if use_backup:
                    uri = self._make_backup_uri(uri)
            self.debug('Loading project from backup: %s', uri)

        # For backup files and legacy formats, force the user to use "Save as"
        if use_backup or path.endswith(".xptv"):
            self.debug("Enforcing read-only mode")
            self.disable_save = True
        else:
            self.disable_save = False

        return uri

    def _is_validate_scenario(self, uri):
        if uri.endswith(".scenario") and has_validate is True:
            # Let's just normally fail if we do not have Validate
            # installed on the system
            return True

        return False

    def _project_pipeline_died_cb(self, unused_pipeline):
        """Shows an dialog telling the user that everything went kaboom."""
        # GTK does not allow an empty string as the dialog title, so we use the
        # same translatable one as render.py's pipeline error message dialog:
        dialog = Gtk.Dialog(title=_("Sorry, something didn’t work right."),
                            transient_for=self.app.gui)

        message = _("Pitivi detected a serious backend problem and could not "
                    "recover from it, even after multiple tries. The only thing "
                    "that can be done at this point is to <b>restart Pitivi</b>."
                    "\n\n"
                    "This is a rare and severe kind of bug. Please see our "
                    "<a href=\"http://developer.pitivi.org/Bug_reporting.html\">"
                    "bug reporting guide</a> and take the time to report it! "
                    "We will be very happy to fix this bug and make sure it "
                    "does not occur again in future versions."
                    "\n\n"
                    "Before closing Pitivi, you can save changes to the "
                    "existing project file or as a separate project file.")

        dialog.add_buttons(_("Save as..."), 1,
                           _("Save"), 2,
                           _("Close Pitivi"), Gtk.ResponseType.CLOSE)

        dialog.set_default_response(1)  # Default to "Save as"
        dialog.set_icon_name("org.pitivi.Pitivi")
        dialog.set_modal(True)
        dialog.get_accessible().set_name("pitivi died")

        primary = Gtk.Label()
        primary.set_line_wrap(True)
        primary.set_use_markup(True)
        primary.set_alignment(0, 0.5)
        primary.props.label = message

        # These 2 lines are needed for a decent dialog width, with wrapped text:
        dialog.props.default_width = 700
        primary.set_width_chars(50)

        # put the text in a vbox
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=SPACING * 2)
        vbox.pack_start(primary, True, True, 0)

        # make the [[image] text] hbox
        image = Gtk.Image.new_from_icon_name("dialog-error-symbolic", Gtk.IconSize.DIALOG)
        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=SPACING * 2)
        hbox.pack_start(image, False, True, 0)
        hbox.pack_start(vbox, True, True, 0)
        hbox.set_border_width(SPACING)

        # stuff the hbox in the dialog
        content_area = dialog.get_content_area()
        content_area.pack_start(hbox, True, True, 0)
        content_area.set_spacing(SPACING * 2)
        hbox.show_all()

        response = dialog.run()
        dialog.destroy()

        if response == 1:
            self.app.gui.editor.save_project_as()
        elif response == 2:
            self.app.gui.editor.save_project()

        self.app.shutdown()

    def load_project(self, uri):
        """Loads the specified URI as a project.

        If a backup file exists, asks if it should be loaded instead, and if so,
        forces the user to use "Save as" afterwards.
        """
        assert self.current_project is None

        is_validate_scenario = self._is_validate_scenario(uri)
        if not is_validate_scenario:
            uri = self._try_using_backup_file(uri)
            scenario = None
        else:
            scenario = path_from_uri(uri)
            uri = None

        # Load the project:
        self.__start_loading_time = time.time()
        project = Project(self.app, uri=uri, scenario=scenario)
        self.emit("new-project-loading", project)

        project.connect_after("missing-uri", self._missing_uri_cb)
        project.connect("loaded", self._project_loaded_cb)

        if not project.create_timeline():
            self.emit("new-project-failed", uri,
                      _('This might be due to a bug or an unsupported project file format. '
                        'If you were trying to add a media file to your project, '
                        'use the "Import" button instead.'))
            return None

        self.current_project = project
        self.emit("new-project-created", project)
        self.current_project.connect("project-changed", self._project_changed_cb)
        self.current_project.pipeline.connect("died", self._project_pipeline_died_cb)

        if is_validate_scenario:
            self.current_project.setup_validate_scenario()

        return project

    def _restore_from_backup_dialog(self, time_diff):
        """Asks if we need to load the autosaved project backup.

        Args:
            time_diff (int): The difference, in seconds, between file mtimes.
        """
        dialog = Gtk.Dialog(title="", transient_for=self.app.gui)
        ignore_backup_btn = dialog.add_button(_("Ignore backup"), Gtk.ResponseType.REJECT)
        ignore_backup_btn.get_style_context().add_class("destructive-action")
        dialog.add_button(_("Restore from backup"), Gtk.ResponseType.YES)
        dialog.set_icon_name("org.pitivi.Pitivi")
        dialog.set_modal(True)
        dialog.set_default_response(Gtk.ResponseType.YES)
        dialog.get_accessible().set_name("restore from backup dialog")

        primary = Gtk.Label()
        primary.set_line_wrap(True)
        primary.set_use_markup(True)
        primary.set_alignment(0, 0.5)

        message = _("An autosaved version of your project file was found. "
                    "It is %s newer than the saved project.\n\n"
                    "Would you like to load it instead?") % \
            beautify_time_delta(time_diff)
        primary.props.label = message

        # put the text in a vbox
        vbox = Gtk.Box(homogeneous=False, spacing=SPACING * 2)
        vbox.set_orientation(Gtk.Orientation.VERTICAL)
        vbox.pack_start(primary, True, True, 0)

        # make the [[image] text] hbox
        image = Gtk.Image.new_from_icon_name(
            "dialog-question-symbolic", Gtk.IconSize.DIALOG)
        hbox = Gtk.Box(homogeneous=False, spacing=SPACING * 2)
        hbox.set_orientation(Gtk.Orientation.HORIZONTAL)
        hbox.pack_start(image, False, False, 0)
        hbox.pack_start(vbox, True, True, 0)
        hbox.set_border_width(SPACING)

        # stuff the hbox in the dialog
        content_area = dialog.get_content_area()
        content_area.pack_start(hbox, True, True, 0)
        content_area.set_spacing(SPACING * 2)
        hbox.show_all()

        response = dialog.run()
        dialog.destroy()
        return response == Gtk.ResponseType.YES

    def save_project(self, uri=None, formatter_type=None, backup=False):
        """Saves the current project.

        Args:
            uri (Optional[str]): If a URI is specified, this means we want to
                save to a new (different) location, so it will be used instead
                of the current project's existing URI.
            formatter_type (Optional[GES.Formatter]): The formatter to use for
                serializing the project. If None, GES defaults to
                GES.XmlFormatter.
            backup (Optional[bool]): Whether to ignore the `uri` arg and save
                the project to a special backup URI built out of the current
                project's URI. Intended for automatic backup behind the scenes.

        Returns:
            bool: Whether the project has been saved successfully.
        """
        if self.disable_save is True and (backup is True or uri is None):
            self.log(
                "Read-only mode is enforced and no new URI was specified, ignoring save request")
            return False

        if backup:
            if self.current_project is not None and self.current_project.uri is not None:
                # Ignore whatever URI that is passed on to us. It's a trap.
                uri = self._make_backup_uri(self.current_project.uri)
            else:
                # Do not try to save backup files for blank projects.
                # It is possible that self.current_project.uri == None when the backup
                # timer sent us an old instance of the (now closed) project.
                return False
        elif uri is None:
            # "Normal save" scenario. The filechoosers in mainwindow ask users
            # for permission to overwrite the file (if needed), so we're safe.
            uri = self.current_project.uri
        else:
            # "Save As" (or "normal-save a blank project") scenario. We use the
            # provided URI, so ensure it's properly encoded, or GIO will fail:
            uri = quote_uri(uri)

            if not is_writable(path_from_uri(uri)):
                # TODO: this will not be needed when GTK+ bug #601451 is fixed
                self.emit("save-project-failed", uri,
                          _("You do not have permissions to write to this folder."))
                return False

        try:
            # "overwrite" is always True: our GTK filechooser save dialogs are
            # set to always ask the user on our behalf about overwriting, so
            # if save_project is actually called, that means overwriting is OK.
            saved = self.current_project.save(
                self.current_project.ges_timeline, uri,
                formatter_type, overwrite=True)
        except GLib.Error as e:
            saved = False
            self.emit("save-project-failed", uri, e)

        if saved:
            if not backup:
                # Do not emit the signal when autosaving a backup file
                self.current_project.set_modification_state(False)
                self.debug('Saved project: %s', uri)
                # Update the project instance's uri,
                # otherwise, subsequent saves will be to the old uri.
                self.info("Setting the project instance's URI to: %s", uri)
                self.current_project.uri = uri
                self.disable_save = False
                self.emit("project-saved", self.current_project, uri)
            else:
                self.debug('Saved backup: %s', uri)

        return saved

    def export_project(self, project, uri):
        """Exports a project and all its media files to a *.tar archive."""
        # Save the project to a temporary file.
        project_name = project.name if project.name else _("project")
        asset = GES.Formatter.get_default()
        project_extension = asset.get_meta(GES.META_FORMATTER_EXTENSION)
        tmp_name = "%s.%s" % (project_name, project_extension)

        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_uri = Gst.filename_to_uri(os.path.join(tmp_dir, tmp_name))
            # save_project updates the project URI... so we better back it up:
            _old_uri = self.current_project.uri
            self.save_project(tmp_uri)
            self.current_project.uri = _old_uri
            # create tar file
            try:
                with tarfile.open(path_from_uri(uri), mode="w") as tar:
                    # top directory in tar-file
                    top = "%s-export" % project_name
                    # add temporary project file
                    tar.add(path_from_uri(tmp_uri), os.path.join(top, tmp_name))

                    # get common path
                    sources = project.list_sources()
                    if self._all_sources_in_homedir(sources):
                        common = os.path.expanduser("~")
                    else:
                        common = "/"

                    # add all sources
                    for source in sources:
                        path = path_from_uri(source.get_id())
                        tar.add(
                            path, os.path.join(top, os.path.relpath(path, common)))
                    tar.close()
            # This catches errors with tarring; the GUI already shows errors while
            # saving projects (ex: permissions), so probably no GUI needed here.
            # Keep the exception generic enough to catch programming errors:
            except tarfile.TarError as e:
                everything_ok = False
                self.error(e)
                tar_file = path_from_uri(uri)
                if os.path.isfile(tar_file):
                    renamed = os.path.splitext(tar_file)[
                        0] + " (CORRUPT)" + "." + project_extension + "_tar"
                    self.warning(
                        'An error occurred, will save the tarball as "%s"', renamed)
                    os.rename(tar_file, renamed)
            else:
                everything_ok = True

        return everything_ok

    def _all_sources_in_homedir(self, sources):
        """Checks if all sources are located in the user's home directory."""
        homedir = os.path.expanduser("~")

        for source in sources:
            if not path_from_uri(source.get_id()).startswith(homedir):
                return False

        return True

    def close_running_project(self):
        """Closes the current project."""
        if self.current_project is None:
            self.warning(
                "Trying to close a project that was already closed/didn't exist")
            return True

        self.info("closing running project %s", self.current_project.uri)
        if not self.emit("closing-project", self.current_project):
            self.warning(
                "Could not close project - this could be because there were unsaved changes and the user cancelled when prompted about them")
            return False

        with Previewer.manager.paused(interrupt=True):
            self.current_project.finalize()

            project = self.current_project
            self.current_project = None
            project.create_thumb()
            self.emit("project-closed", project)
            disconnect_all_by_func(project, self._project_changed_cb)
            disconnect_all_by_func(project.pipeline, self._project_pipeline_died_cb)
            self._clean_backup(project.uri)
            self.exitcode = project.release()

        return True

    def new_blank_project(self):
        """Creates a new blank project and sets it as the current project.

        Returns:
            Project: The created project.
        """
        self.debug("New blank project")

        assert self.current_project is None

        self.__start_loading_time = time.time()
        project = Project(self.app)
        self.emit("new-project-loading", project)

        # setting default values for project metadata
        project.author = pwd.getpwuid(os.getuid()).pw_gecos.split(",")[0]

        project.create_timeline()
        project._ensure_tracks()  # pylint: disable=protected-access
        project.update_restriction_caps()
        self.current_project = project
        self.emit("new-project-created", project)

        project.connect("project-changed", self._project_changed_cb)
        project.pipeline.connect("died", self._project_pipeline_died_cb)
        project.set_modification_state(False)
        self.emit("new-project-loaded", self.current_project)
        project.loaded = True
        self.time_loaded = time.time()

        return project

    def revert_to_saved_project(self):
        """Discards all unsaved changes and reloads the current open project."""
        if self.current_project.uri is None or not self.current_project.has_unsaved_modifications():
            return True
        if not self.emit("reverting-to-saved", self.current_project):
            return False

        uri = self.current_project.uri
        self.current_project.set_modification_state(False)
        self.close_running_project()
        self.load_project(uri)
        return True

    def _project_changed_cb(self, project):
        # _backup_lock is a timer, when a change in the project is done it is
        # set to 10 seconds. If before those 10 secs pass another change occurs,
        # 5 secs are added to the timeout callback instead of saving the backup
        # file. The limit is 60 seconds.
        uri = project.uri
        if uri is None:
            return

        if self._backup_lock == 0:
            self._backup_lock = 10
            GLib.timeout_add_seconds(
                self._backup_lock, self._save_backup_cb, project, uri)
        else:
            if self._backup_lock < 60:
                self._backup_lock += 5

    def _save_backup_cb(self, unused_project, unused_uri):
        if self._backup_lock > 10:
            self._backup_lock -= 5
            return True
        else:
            self.save_project(backup=True)
            self._backup_lock = 0
        return False

    def _clean_backup(self, uri):
        if uri is None:
            return
        path = path_from_uri(self._make_backup_uri(uri))
        if os.path.exists(path):
            os.remove(path)
            self.debug('Removed backup file: %s', path)

    def _make_backup_uri(self, uri):
        """Generates a corresponding backup URI or path.

        This does not guarantee that the backup file actually exists or that
        the file extension is actually a project file.

        Args:
            uri (str): The project URI or file path.

        Returns:
            str: The backup version of the `uri`.
        """
        name, ext = os.path.splitext(uri)
        return name + ext + "~"

    def _missing_uri_cb(self, project, error, asset):
        new_uri = self.emit("missing-uri", project, error, asset)
        if not new_uri:
            project.at_least_one_asset_missing = True
        else:
            project.relocated_assets[asset.props.id] = new_uri
        project.set_modification_state(True)
        return new_uri

    def _project_loaded_cb(self, project, unused_timeline):
        self.debug("Project loaded %s", project.props.uri)
        if not self.current_project == project:
            self.debug("Project is obsolete %s", project.props.uri)
            return
        self.emit("new-project-loaded", project)
        project.loaded = True
        self.time_loaded = time.time()
        self.info("Loaded in %s", self.time_loaded - self.__start_loading_time)


class Project(Loggable, GES.Project):
    """A Pitivi project.

    Attributes:
        app (Pitivi): The app.
        name (str): The name of the project.
        description (str): The description of the project.
        ges_timeline (GES.Timeline): The timeline.
        pipeline (Pipeline): The timeline's pipeline.
        loaded (bool): Whether the project is fully loaded.

    Args:
        uri (Optional[str]): The URI of the file where the project should
            be loaded from.

    Signals:
        project-changed: Modifications were made to the project.
        start-importing: Started to import files.
    """

    __gsignals__ = {
        "asset-loading-progress": (GObject.SignalFlags.RUN_LAST, None, (object, int)),
        # Working around the fact that PyGObject does not let us emit error-loading-asset
        # and bugzilla does not let me file a bug right now :/
        "proxying-error": (GObject.SignalFlags.RUN_LAST, None,
                           (object,)),
        "start-importing": (GObject.SignalFlags.RUN_LAST, None, ()),
        "project-changed": (GObject.SignalFlags.RUN_LAST, None, ()),
        "settings-set-from-imported-asset": (GObject.SignalFlags.RUN_LAST, None,
                                             (GES.Asset,)),
        "video-size-changed": (GObject.SignalFlags.RUN_LAST, None, ()),
        "audio-channels-changed": (GObject.SignalFlags.RUN_LAST, None, ()),
        "safe-area-size-changed": (GObject.SignalFlags.RUN_LAST, None, ())
    }

    def __init__(self, app, uri=None, scenario=None, **unused_kwargs):
        Loggable.__init__(self)
        GES.Project.__init__(self, uri=uri, extractable_type=GES.Timeline)
        self.log("uri:%s", uri)
        self.pipeline = None
        self.ges_timeline = None
        self.uri = uri
        self.loaded = False
        self.at_least_one_asset_missing = False
        self.app = app
        self.loading_assets = set()

        self.relocated_assets = {}
        self.app.proxy_manager.connect("progress", self.__asset_transcoding_progress_cb)
        self.app.proxy_manager.connect("error-preparing-asset",
                                       self.__proxy_error_cb)
        self.app.proxy_manager.connect("asset-preparing-cancelled",
                                       self.__asset_transcoding_cancelled_cb)
        self.app.proxy_manager.connect("proxy-ready",
                                       self.__proxy_ready_cb)

        # GstValidate
        self.scenario = scenario
        self.runner = None
        self.monitor = None
        self._scenario = None

        # For keeping track of assets importing.
        self._dirty = False
        self.nb_remaining_file_to_import = 0
        self.nb_imported_files = 0

        # Main assets that were proxied when saving the project but
        # whose proxies had been deleted from the filesystem. The
        # proxy files are being regenerated.
        self.__deleted_proxy_files = set()

        # List of proxy assets uris that were deleted on the filesystem
        # and we are waiting for the main asset (ie. the file from
        # which the proxy was generated) to be loaded before we can try to
        # regenerate the proxy.
        self.__awaited_deleted_proxy_targets = set()

        # Project property default values
        self.register_meta(GES.MetaFlag.READWRITE, "author", "")

        self.register_meta(GES.MetaFlag.READWRITE, "scaled_proxy_width", 0)
        self.register_meta(GES.MetaFlag.READWRITE, "scaled_proxy_height", 0)
        self.register_meta(GES.MetaFlag.READWRITE, "pitivi::title_safe_area_vertical", DEFAULT_TITLE_AREA_VERTICAL)
        self.register_meta(GES.MetaFlag.READWRITE, "pitivi::title_safe_area_horizontal", DEFAULT_TITLE_AREA_HORIZONTAL)
        self.register_meta(GES.MetaFlag.READWRITE, "pitivi::action_safe_area_vertical", DEFAULT_ACTION_AREA_VERTICAL)
        self.register_meta(GES.MetaFlag.READWRITE, "pitivi::action_safe_area_horizontal", DEFAULT_ACTION_AREA_HORIZONTAL)

        # The rendering settings.
        self.set_meta("render-scale", 100.0)

        self.container_profile = \
            GstPbutils.EncodingContainerProfile.new("pitivi-profile",
                                                    _("Pitivi encoding profile"),
                                                    Gst.Caps("video/webm"),
                                                    None)
        has_default_settings = not bool(uri) and not bool(scenario)
        vsettings = DEFAULT_VIDEO_SETTINGS if has_default_settings else "video/x-raw"
        self.video_profile = GstPbutils.EncodingVideoProfile.new(
            Gst.Caps("video/x-vp8"), None, Gst.Caps(vsettings), 0)
        asettings = DEFAULT_AUDIO_SETTINGS if has_default_settings else "audio/x-raw"
        self.audio_profile = GstPbutils.EncodingAudioProfile.new(
            Gst.Caps("audio/x-vorbis"), None, Gst.Caps(asettings), 0)
        self.container_profile.add_profile(self.video_profile)
        self.container_profile.add_profile(self.audio_profile)

        # Add the container profile to the project so it is saved
        # as part of the project.
        res = self.add_encoding_profile(self.container_profile)
        assert res

        self.muxer = Encoders().default_muxer
        self.vencoder = Encoders().default_video_encoder
        self.aencoder = Encoders().default_audio_encoder
        res = self._ensure_audio_restrictions()
        assert res
        res = self._ensure_video_restrictions()
        assert res
        self._has_default_audio_settings = has_default_settings
        self._has_default_video_settings = has_default_settings

        # A ((container_profile, muxer) -> containersettings) map.
        self._containersettings_cache = {}
        # A ((container_profile, vencoder) -> vcodecsettings) map.
        self._vcodecsettings_cache = {}
        # A ((container_profile, aencoder) -> acodecsettings) map.
        self._acodecsettings_cache = {}
        # Whether the current settings are temporary and should be reverted,
        # as they apply only for rendering.
        self._has_rendering_values = False

    def _scenario_done_cb(self, scenario):
        if self.pipeline is not None:
            self.pipeline.set_force_position_listener(False)

    def setup_validate_scenario(self):
        from gi.repository import GstValidate

        self.info("Setting up validate scenario")
        self.runner = GstValidate.Runner.new()
        create_monitor(self.runner, self.app.gui)
        self.monitor = GstValidate.Monitor.factory_create(
            self.pipeline, self.runner, None)
        self._scenario = GstValidate.Scenario.factory_create(
            self.runner, self.pipeline, self.scenario)
        self.pipeline.set_force_position_listener(True)
        self._scenario.connect("done", self._scenario_done_cb)
        self._scenario.props.execute_on_idle = True

    # --------------- #
    # Our properties  #
    # --------------- #

    # Project specific properties
    @property
    def name(self):
        if not self.uri:
            return DEFAULT_NAME
        quoted_name = os.path.splitext(os.path.basename(self.uri))[0]
        return unquote(quoted_name)

    @property
    def year(self):
        return self.get_meta("year")

    @year.setter
    def year(self, year):
        if year == self.year:
            return
        self.set_meta("year", year)

    @property
    def description(self):
        return self.get_meta("description")

    @description.setter
    def description(self, description):
        if description == self.description:
            return
        self.set_meta("description", description)

    @property
    def author(self):
        return self.get_meta("author")

    @author.setter
    def author(self, author):
        if author == self.author:
            return
        self.set_meta("author", author)

    @property
    def scaled_proxy_height(self):
        return self.get_meta("scaled_proxy_height") or self.app.settings.default_scaled_proxy_height

    @scaled_proxy_height.setter
    def scaled_proxy_height(self, scaled_proxy_height):
        if scaled_proxy_height == self.get_meta("scaled_proxy_height"):
            return
        self.set_meta("scaled_proxy_height", scaled_proxy_height)
        self.set_modification_state(True)

    @property
    def scaled_proxy_width(self):
        return self.get_meta("scaled_proxy_width") or self.app.settings.default_scaled_proxy_width

    @scaled_proxy_width.setter
    def scaled_proxy_width(self, scaled_proxy_width):
        if scaled_proxy_width == self.get_meta("scaled_proxy_width"):
            return
        self.set_meta("scaled_proxy_width", scaled_proxy_width)
        self.set_modification_state(True)

    def has_scaled_proxy_size(self):
        """Returns whether the proxy size has been set."""
        return bool(self.get_meta("scaled_proxy_width") and self.get_meta("scaled_proxy_height"))

    @staticmethod
    def get_thumb_path(uri, resolution):
        """Returns path of thumbnail of specified resolution in the cache."""
        thumb_hash = md5(quote_uri(uri).encode()).hexdigest()
        thumbs_cache_dir = xdg_cache_home("project_thumbs", resolution)
        return os.path.join(thumbs_cache_dir, thumb_hash) + ".png"

    @classmethod
    def get_thumb(cls, uri):
        """Gets the project thumb, if exists, else the default thumb or None."""
        try:
            thumb = GdkPixbuf.Pixbuf.new_from_file(cls.get_thumb_path(uri, SCALED_THUMB_DIR))
        except GLib.Error:
            # Try to get the default thumb.
            try:
                thumb = Gtk.IconTheme.get_default().load_icon("video-x-generic", 128, 0)
            except GLib.Error:
                return None
            thumb = scale_pixbuf(thumb, SCALED_THUMB_WIDTH, SCALED_THUMB_HEIGHT)

        return thumb

    def __create_scaled_thumb(self):
        """Creates scaled thumbnail from the original thumbnail."""
        try:
            thumb = GdkPixbuf.Pixbuf.new_from_file(self.get_thumb_path(self.uri, ORIGINAL_THUMB_DIR))
            thumb = scale_pixbuf(thumb, SCALED_THUMB_WIDTH, SCALED_THUMB_HEIGHT)
            thumb.savev(self.get_thumb_path(self.uri, SCALED_THUMB_DIR), "png", [], [])
        except GLib.Error as e:
            self.warning("Failed to create scaled project thumbnail: %s", e)

    def __remove_thumbs(self):
        """Removes existing project thumbnails."""
        for thumb_dir in (ORIGINAL_THUMB_DIR, SCALED_THUMB_DIR):
            try:
                os.remove(self.get_thumb_path(self.uri, thumb_dir))
            except FileNotFoundError:
                pass

    @staticmethod
    def __pick_thumb_from_assets_thumbs(assets):
        """Picks project thumbnail from assets thumbnails."""
        for asset in assets:
            thumb_cache = ThumbnailCache.get(asset)
            thumb = thumb_cache.get_preview_thumbnail()
            if thumb:
                # First asset that has a preview thumbnail.
                return thumb
        return None

    def create_thumb(self):
        """Creates project thumbnails."""
        if not self.uri:
            return

        thumb_path = self.get_thumb_path(self.uri, ORIGINAL_THUMB_DIR)

        if os.path.exists(thumb_path) and not self.app.action_log.has_assets_operations():
            # The project thumbnail already exists and the assets are the same.
            return

        # Project Thumbnail Generation Approach: Out of thumbnails of all
        # the assets in the current project, the one with maximum file size
        # will be our project thumbnail - http://bit.ly/thumbnail-generation

        assets = self.list_sources()
        assets_uri = [asset.props.id for asset in assets]

        if not assets_uri:
            # There are no assets in the project,
            # so make sure there are no project thumbs.
            self.__remove_thumbs()
            return

        normal_thumb_path = None
        large_thumb_path = None
        normal_thumb_size = 0
        large_thumb_size = 0
        n_normal_thumbs = 0
        n_large_thumbs = 0

        for uri in assets_uri:
            path_128, path_256 = AssetThumbnail.get_asset_thumbnails_path(uri)

            try:
                thumb_size = os.stat(path_128).st_size
                if thumb_size > normal_thumb_size:
                    normal_thumb_path = path_128
                    normal_thumb_size = thumb_size
                n_normal_thumbs += 1
            except FileNotFoundError:
                # The asset is missing the normal thumbnail.
                pass

            try:
                thumb_size = os.stat(path_256).st_size
                if thumb_size > large_thumb_size:
                    large_thumb_path = path_256
                    large_thumb_size = thumb_size
                n_large_thumbs += 1
            except FileNotFoundError:
                # The asset is missing the large thumbnail.
                pass

        if normal_thumb_path or large_thumb_path:
            # Use the category for which we found the max number of
            # thumbnails to find the most complex thumbnail, because
            # we can't compare the small with the large.
            if n_normal_thumbs > n_large_thumbs:
                shutil.copyfile(normal_thumb_path, thumb_path)
            else:
                shutil.copyfile(large_thumb_path, thumb_path)
        else:
            # No thumbnails available in the XDG cache.
            thumb = self.__pick_thumb_from_assets_thumbs(assets)
            if not thumb:
                return
            thumb.savev(thumb_path, "png", [], [])

        self.__create_scaled_thumb()

    def set_rendering(self, rendering):
        """Sets the a/v restrictions for rendering or for editing."""
        res = self._ensure_audio_restrictions()
        assert res
        res = self._ensure_video_restrictions()
        assert res

        video_restrictions = self.video_profile.get_restriction().copy_nth(0)

        if self._has_rendering_values != rendering:
            # pylint: disable=attribute-defined-outside-init
            if rendering:
                video_restrictions_struct = video_restrictions[0]
                self.__width = video_restrictions_struct["width"]
                self.__height = video_restrictions_struct["height"]
                width = int(self.__width * self.render_scale / 100)
                height = int(self.__height * self.render_scale / 100)
            else:
                width = self.__width
                height = self.__height

            video_restrictions.set_value("width", width)
            video_restrictions.set_value("height", height)

        self._has_rendering_values = rendering
        self.video_profile.set_restriction(video_restrictions)

    @staticmethod
    def _set_restriction(profile, name, value):
        """Sets a restriction on the specified profile.

        Assumes the profile has a single restriction.

        Args:
            profile (GstPbutils.EncodingProfile): The profile to be updated.

        Returns:
            bool: Whether the profile actually changed.
        """
        caps = profile.get_restriction()
        if caps[0][name] == value or not value:
            return False

        restriction = caps.copy_nth(0)
        restriction.set_value(name, value)
        profile.set_restriction(restriction)
        return True

    def _set_video_restriction(self, name, value):
        """Updates the video profile and the corresponding project settings."""
        res = Project._set_restriction(self.video_profile, name, value)
        if res:
            self.emit("video-size-changed")
            self._has_default_video_settings = False
        return res

    def _set_audio_restriction(self, name, value):
        """Updates the audio profile and the corresponding project settings."""
        res = Project._set_restriction(self.audio_profile, name, value)
        if res:
            self.emit("audio-channels-changed")
            self._has_default_audio_settings = False
            self.update_restriction_caps()
        return res

    @property
    def videowidth(self):
        return self.video_profile.get_restriction()[0]["width"]

    @videowidth.setter
    def videowidth(self, value):
        if self._set_video_restriction("width", int(value)):
            self.update_restriction_caps()
            self.set_modification_state(True)

    @property
    def videoheight(self):
        return self.video_profile.get_restriction()[0]["height"]

    @videoheight.setter
    def videoheight(self, value):
        if self._set_video_restriction("height", int(value)):
            self.update_restriction_caps()
            self.set_modification_state(True)

    @property
    def videorate(self):
        return self.video_profile.get_restriction()[0]["framerate"]

    @videorate.setter
    def videorate(self, value):
        if self._set_video_restriction("framerate", value):
            self.update_restriction_caps()
            self.set_modification_state(True)

    def set_video_properties(self, width, height, framerate):
        """Sets the video properties in one operation.

        This should be called when several properties can be changed at once,
        to avoid GES repositioning all sources when the video size changes.

        Args:
            width (int): The new project width.
            height (int): The new project height.
            framerate (Gst.Fraction): The new project framerate.
        """
        changed = any([self._set_video_restriction("width", int(width)),
                       self._set_video_restriction("height", int(height)),
                       self._set_video_restriction("framerate", framerate)])
        if changed:
            self.update_restriction_caps()
            self.set_modification_state(True)

    @property
    def audiochannels(self) -> int:
        # The map does not always contain "channels".
        return self.audio_profile.get_restriction()[0]["channels"] or 0

    @audiochannels.setter
    def audiochannels(self, value: int):
        if self._set_audio_restriction("channels", int(value)):
            self.set_modification_state(True)

    @property
    def audiorate(self):
        try:
            return int(self.audio_profile.get_restriction()[0]["rate"])
        except TypeError:
            return None

    @audiorate.setter
    def audiorate(self, value):
        if self._set_audio_restriction("rate", int(value)):
            self.set_modification_state(True)

    @property
    def aencoder(self):
        return self.audio_profile.get_preset_name()

    @aencoder.setter
    def aencoder(self, preset_factory_name):
        if self._update_encoding_profile(self.audio_profile, preset_factory_name):
            self.set_modification_state(True)

    @property
    def vencoder(self):
        return self.video_profile.get_preset_name()

    @vencoder.setter
    def vencoder(self, preset_factory_name):
        if self._update_encoding_profile(self.video_profile, preset_factory_name):
            self.set_modification_state(True)

    @property
    def muxer(self):
        return self.container_profile.get_preset_name()

    @muxer.setter
    def muxer(self, preset_factory_name):
        if self._update_encoding_profile(self.container_profile, preset_factory_name):
            self.set_modification_state(True)

    def _update_encoding_profile(self, profile, preset_factory_name):
        """Updates the specified encoding profile.

        Args:
            profile (GstPbutils.EncodingProfile): The profile to be updated.
            preset_factory_name (str): The name of the Gst.Preset’s factory
                to be used in the profile.

        Returns:
            bool: Whether the profile has been changed.
        """
        if profile.get_preset_name() == preset_factory_name or not preset_factory_name:
            return False

        caps = self._get_caps_from_feature(preset_factory_name)
        if caps:
            profile.set_format(caps)

        # Set the name of the factory for producing the audio encoder.
        profile.set_preset_name(preset_factory_name)

        # Make sure the encoder does not use any encoding preset.
        # Gst.Preset can be set exclusively through EncodingTargets for now.
        profile.set_preset(None)

        return True

    def _get_caps_from_feature(self, name):
        """Gets the caps for the source static pad template of a feature."""
        feature = Gst.Registry.get().lookup_feature(name)
        if not feature:
            self.error("%s not in registry", name)
            return None
        for template in feature.get_static_pad_templates():
            if template.name_template == "src":
                return Gst.Caps(template.get_caps()[0].to_string())
        self.error("%s has no source static pad templates", name)
        return None

    @property
    def render_scale(self):
        return self.get_meta("render-scale")

    @render_scale.setter
    def render_scale(self, value):
        if value:
            self.set_meta("render-scale", value)

    def set_safe_areas_sizes(self, title_horizontal_factor, title_vertical_factor, action_horizontal_factor, action_vertical_factor):
        """Sets the safe areas sizes in one operation."""
        self.title_safe_area_horizontal = title_horizontal_factor
        self.title_safe_area_vertical = title_vertical_factor
        self.action_safe_area_horizontal = action_horizontal_factor
        self.action_safe_area_vertical = action_vertical_factor

    @property
    def title_safe_area_vertical(self):
        return self.get_meta("pitivi::title_safe_area_vertical")

    @title_safe_area_vertical.setter
    def title_safe_area_vertical(self, percentage):
        if percentage == self.get_meta("pitivi::title_safe_area_vertical"):
            return
        self.set_meta("pitivi::title_safe_area_vertical", percentage)
        self.emit("safe-area-size-changed")

    @property
    def title_safe_area_horizontal(self):
        return self.get_meta("pitivi::title_safe_area_horizontal")

    @title_safe_area_horizontal.setter
    def title_safe_area_horizontal(self, percentage):
        if percentage == self.get_meta("pitivi::title_safe_area_horizontal"):
            return
        self.set_meta("pitivi::title_safe_area_horizontal", percentage)
        self.emit("safe-area-size-changed")

    @property
    def action_safe_area_vertical(self):
        return self.get_meta("pitivi::action_safe_area_vertical")

    @action_safe_area_vertical.setter
    def action_safe_area_vertical(self, percentage):
        if percentage == self.get_meta("pitivi::action_safe_area_vertical"):
            return
        self.set_meta("pitivi::action_safe_area_vertical", percentage)
        self.emit("safe-area-size-changed")

    @property
    def action_safe_area_horizontal(self):
        return self.get_meta("pitivi::action_safe_area_horizontal")

    @action_safe_area_horizontal.setter
    def action_safe_area_horizontal(self, percentage):
        if percentage == self.get_meta("pitivi::action_safe_area_horizontal"):
            return
        self.set_meta("pitivi::action_safe_area_horizontal", percentage)
        self.emit("safe-area-size-changed")

    # ------------------------------#
    # Proxy creation implementation #
    # ------------------------------#
    def __asset_transcoding_progress_cb(self, proxy_manager, asset,
                                        creation_progress, estimated_time):
        self.__update_asset_loading_progress(estimated_time)

    def __get_loading_project_progress(self):
        """Computes current advancement of asset loading during project loading.

        During project loading we keep all loading assets to keep track of real advancement
        during the whole process, whereas while adding new assets, they get removed from
        the `loading_assets` list once the proxy is ready.

        Returns:
            int: The current asset loading progress (in percent).
        """
        num_loaded = 0
        all_ready = True
        for asset in self.loading_assets:
            if asset.creation_progress < 100:
                all_ready = False
            else:
                # Check that we are not recreating deleted proxy
                proxy_uri = self.app.proxy_manager.get_proxy_uri(asset)
                scaled_proxy_uri = self.app.proxy_manager.get_proxy_uri(asset, scaled=True)

                no_hq_proxy = False
                no_scaled_proxy = False

                if proxy_uri and proxy_uri not in self.__deleted_proxy_files and \
                        asset.props.id not in self.__awaited_deleted_proxy_targets:
                    no_hq_proxy = True

                if scaled_proxy_uri and scaled_proxy_uri not in self.__deleted_proxy_files and \
                        asset.props.id not in self.__awaited_deleted_proxy_targets:
                    no_scaled_proxy = True

                if no_hq_proxy and no_scaled_proxy:
                    asset.ready = True

                num_loaded += 1

        if all_ready:
            return 100

        return (num_loaded / len(self.loading_assets)) * 100

    def __get_loading_assets_progress(self):
        """Computes current advancement of asset loading.

        Returns:
            int: The current asset loading progress (in percent).
        """
        total_import_duration = 0
        for asset in self.loading_assets:
            total_import_duration += asset.get_duration()

        if total_import_duration == 0:
            self.info("No known duration yet")
            return 0

        asset_loading_progress = 0
        all_ready = True
        for asset in self.loading_assets:
            asset_weight = asset.get_duration() / total_import_duration
            asset_loading_progress += asset_weight * asset.creation_progress

            if asset.creation_progress < 100:
                all_ready = False
            elif not asset.ready:
                self.set_modification_state(True)
                asset.ready = True

        if all_ready:
            asset_loading_progress = 100

        return asset_loading_progress

    def __update_asset_loading_progress(self, estimated_time=0):
        if not self.loading_assets:
            self.emit("asset-loading-progress", 100, estimated_time)
            return

        if not self.loaded:
            progress = self.__get_loading_project_progress()
        else:
            progress = self.__get_loading_assets_progress()

        self.emit("asset-loading-progress", progress, estimated_time)

        if progress == 100:
            self.info("No more loading assets")
            self.loading_assets = set()

    def __asset_transcoding_cancelled_cb(self, unused_proxy_manager, asset):
        self.__set_proxy(asset, None)
        self.__update_asset_loading_progress()

    def __proxy_error_cb(self, unused_proxy_manager, asset, proxy, error):
        if asset is None:
            asset_id = self.app.proxy_manager.get_target_uri(proxy)
            if asset_id:
                asset = GES.Asset.request(proxy.get_extractable_type(),
                                          asset_id)
                if not asset:
                    for tmpasset in self.loading_assets:
                        if tmpasset.props.id == asset_id:
                            asset = tmpasset
                            break

                    if not asset:
                        self.error("Could not get the asset %s from its proxy %s", asset_id,
                                   proxy.props.id)

                        return
            else:
                self.info("%s is not a proxy asset", proxy.props.id)

                return

        asset.proxying_error = error
        asset.creation_progress = 100

        self.emit("proxying-error", asset)
        self.__update_asset_loading_progress()

    def __proxy_ready_cb(self, unused_proxy_manager, asset, proxy):
        if proxy and proxy.props.id in self.__deleted_proxy_files:
            self.info("Recreated proxy is now ready, stop having"
                      " its target as a proxy.")
            proxy.unproxy(asset)

        self.__set_proxy(asset, proxy)

    def __set_proxy(self, asset, proxy):
        asset.creation_progress = 100
        asset.ready = True
        if proxy:
            self.finalize_proxy(proxy)

        asset.set_proxy(proxy)
        try:
            self.loading_assets.remove(asset)
        except KeyError:
            pass

        if proxy:
            self.add_asset(proxy)
            self.loading_assets.add(proxy)

        self.__update_asset_loading_progress()

    def finalize_proxy(self, proxy):
        proxy.ready = False
        proxy.error = None
        proxy.creation_progress = 100

    # ------------------------------------------ #
    # GES.Project virtual methods implementation #
    # ------------------------------------------ #
    def do_asset_loading(self, asset):
        if asset and not GObject.type_is_a(asset.get_extractable_type(), GES.UriClip):
            # Ignore for example the assets producing GES.TitleClips.
            return

        self._prepare_asset_processing(asset)

    def __regenerate_missing_proxy(self, asset, scaled=False):
        self.info("Re generating deleted proxy file %s.", asset.props.id)
        GES.Asset.needs_reload(GES.UriClip, asset.props.id)
        self._prepare_asset_processing(asset)
        asset.force_proxying = True
        self.app.proxy_manager.add_job(asset, scaled=scaled)
        self.__update_asset_loading_progress()

    def do_missing_uri(self, error, asset):
        if self.app.proxy_manager.is_proxy_asset(asset):
            self.debug("Missing proxy file: %s", asset.props.id)
            target_uri = self.app.proxy_manager.get_target_uri(asset)

            # Take asset relocation into account.:
            target_uri = self.relocated_assets.get(target_uri, target_uri)

            GES.Asset.needs_reload(GES.UriClip, asset.props.id)
            # Check if the target has already been loaded.
            target = [asset for asset in self.list_assets(GES.UriClip) if
                      asset.props.id == target_uri]
            if target:
                scaled = self.app.proxy_manager.is_scaled_proxy(asset)
                self.__regenerate_missing_proxy(target[0], scaled=scaled)
            else:
                self.__awaited_deleted_proxy_targets.add(target_uri)

            self.__deleted_proxy_files.add(asset.props.id)
            return target_uri

        new_uri = GES.Project.do_missing_uri(self, error, asset)

        if new_uri:
            self.relocated_assets[asset.props.id] = new_uri

        return new_uri

    def _prepare_asset_processing(self, asset):
        asset.creation_progress = 0
        asset.error = None
        asset.ready = False
        asset.force_proxying = False
        asset.proxying_error = None

        if not self.loading_assets:
            # Progress == 0 means "starting to import"
            self.emit("asset-loading-progress", 0, 0)

        self.loading_assets.add(asset)

    def do_asset_removed(self, asset):
        self.app.proxy_manager.cancel_job(asset)

    def do_asset_added(self, asset):
        """Handles `GES.Project::asset-added` emitted by self."""
        self._maybe_init_settings_from_asset(asset)
        if asset and not GObject.type_is_a(asset.get_extractable_type(),
                                           GES.UriClip):
            # Ignore for example the assets producing GES.TitleClips.
            self.debug("Ignoring asset: %s", asset.props.id)
            return

        if asset not in self.loading_assets:
            self.debug("Asset %s is not in loading assets, "
                       " it must not be proxied", asset.get_id())
            return

        if asset.props.id in self.__awaited_deleted_proxy_targets:
            self.__regenerate_missing_proxy(asset)
            self.__awaited_deleted_proxy_targets.remove(asset.props.id)
        elif asset.props.id in self.__deleted_proxy_files:
            self.info("Deleted proxy file %s now ready again.", asset.props.id)
            self.__deleted_proxy_files.remove(asset.props.id)

        if self.loaded:
            if not asset.get_proxy_target() in self.list_assets(GES.Extractable):
                self.app.proxy_manager.add_job(asset)
        else:
            self.debug("Project still loading, not using proxies: %s",
                       asset.props.id)
            asset.creation_progress = 100
            self.__update_asset_loading_progress()

    def do_loading_error(self, error, asset_id, unused_type):
        """Handles `GES.Project::error-loading-asset` emitted by self."""
        asset = None
        for asset in self.loading_assets:
            if asset.get_id() == asset_id:
                break

        self.error("Could not load %s: %s -> %s", asset_id, error, asset)
        asset.error = error
        asset.creation_progress = 100
        if self.loaded:
            self.loading_assets.remove(asset)
        self.__update_asset_loading_progress()

    def do_loaded(self, unused_timeline):
        """Handles `GES.Project::loaded` emitted by self."""
        if not self.ges_timeline:
            return

        self._ensure_tracks()
        self.ges_timeline.props.auto_transition = True
        self._ensure_layer()

        if self.uri:
            self.loading_assets = {asset for asset in self.loading_assets
                                   if self.app.proxy_manager.is_asset_queued(asset)}

            if self.loading_assets:
                self.debug("The following assets are still being transcoded: %s."
                           " (They must be proxied assets with missing/deleted"
                           " proxy files).", self.loading_assets)
            self.__update_asset_loading_progress()

        if self.scenario is not None:
            return

        profiles = self.list_encoding_profiles()
        # The project just loaded, check the new
        # encoding profile and make use of it now.
        self.info("Using first usable encoding profile: %s", [p.get_preset_name() for p in profiles])
        for profile in profiles:
            if self.set_container_profile(profile):
                break

        self._load_encoder_settings(profiles)

    def set_container_profile(self, container_profile: GstPbutils.EncodingContainerProfile) -> bool:
        """Sets the specified container profile as new profile if usable.

        Returns:
            True if it has been set successfully.
        """
        if container_profile == self.container_profile:
            return True

        muxer = self.get_element_factory_name(container_profile)
        if muxer is None:
            muxer = Encoders().default_muxer
        container_profile.set_preset_name(muxer)

        video_profile: Optional[GstPbutils.EncodingVideoProfile] = None
        audio_profile: Optional[GstPbutils.EncodingAudioProfile] = None
        vencoder: Optional[str] = None
        aencoder: Optional[str] = None
        for profile in container_profile.get_profiles():
            if isinstance(profile, GstPbutils.EncodingVideoProfile):
                video_profile = profile
                if profile.get_restriction() is None:
                    profile.set_restriction(Gst.Caps("video/x-raw"))

                if not self._ensure_video_restrictions(profile):
                    return False

                vencoder = self.get_element_factory_name(profile)
                if vencoder:
                    profile.set_preset_name(vencoder)
            elif isinstance(profile, GstPbutils.EncodingAudioProfile):
                audio_profile = profile
                if profile.get_restriction() is None:
                    profile.set_restriction(Gst.Caps("audio/x-raw"))

                if not self._ensure_audio_restrictions(profile):
                    return False

                aencoder = self.get_element_factory_name(profile)
                if aencoder:
                    profile.set_preset_name(aencoder)
            else:
                self.warning("We do not handle profile: %s", profile)

        if not aencoder:
            self.error("Can't use profile, no audio encoder found.")
            return False

        if not vencoder:
            self.error("Can't use profile, no video encoder found.")
            return False

        self.container_profile = container_profile
        self.video_profile = video_profile
        self.audio_profile = audio_profile

        return True

    def is_profile_subset(self, profile, superset):
        return self.get_element_factory_name(profile) == self.get_element_factory_name(superset)

    def matches_container_profile(self, container_profile):
        if not self.is_profile_subset(container_profile, self.container_profile):
            return False

        video_matches = False
        has_video = False
        audio_matches = False
        has_audio = False
        for profile in container_profile.get_profiles():
            if isinstance(profile, GstPbutils.EncodingVideoProfile):
                has_video = True
                video_matches |= self.is_profile_subset(profile, self.video_profile)

                # For example: "Profile Youtube"
                preset_name = profile.get_preset()
                if preset_name:
                    # We assume container_profile has the same preset
                    # as the included video profile.

                    current_preset_name = self.video_profile.get_preset()
                    if not current_preset_name:
                        return False

                    # For example: "x264enc"
                    preset_factory_name = self.video_profile.get_preset_name()
                    tmp_preset = Gst.ElementFactory.make(preset_factory_name, None)
                    tmp_preset.load_preset(current_preset_name)

                    res, last_applied_preset_name = tmp_preset.get_meta(current_preset_name, "pitivi::OriginalPreset")
                    if res and preset_name != last_applied_preset_name:
                        return False

            elif isinstance(profile, GstPbutils.EncodingAudioProfile):
                has_audio = True
                audio_matches |= self.is_profile_subset(profile, self.audio_profile)

        if has_audio:
            if not audio_matches:
                return False

        if has_video:
            if not video_matches:
                return False

        return True

    def _load_encoder_settings(self, profiles):
        for container_profile in profiles:
            if not isinstance(container_profile, GstPbutils.EncodingContainerProfile):
                self.warning("%s is not an EncodingContainerProfile", container_profile)
                continue

            for profile in container_profile.get_profiles():
                preset_name = profile.get_preset()
                if not preset_name:
                    continue

                preset_factory_name = profile.get_preset_name()
                preset = Gst.ElementFactory.make(preset_factory_name, None)
                if not isinstance(preset, Gst.Preset):
                    self.warning("Element %s does not implement Gst.Preset. Cannot load"
                                 "its rendering settings", preset)
                    continue

                if profile.get_type_nick() == "video":
                    cache = self._vcodecsettings_cache
                elif profile.get_type_nick() == "audio":
                    cache = self._acodecsettings_cache
                else:
                    self.warning("Unrecognized profile type for profile %s", profile)
                    continue
                cache_key = (container_profile, preset_factory_name)

                if not preset.load_preset(preset_name):
                    self.warning("No preset named %s for encoder %s", preset_name, preset)
                    continue

                cache[cache_key] = {prop.name: preset.get_property(prop.name)
                                    for prop in GObject.list_properties(preset)
                                    if prop.name not in IGNORED_PROPS and prop.flags & GObject.ParamFlags.WRITABLE}

    # ------------------------------------------ #
    # Our API                                    #
    # ------------------------------------------ #

    def finalize(self):
        """Disconnects all signals and everything.

        Makes sure the project won't be doing anything after the call.
        """
        if self._scenario:
            self._scenario.disconnect_by_func(self._scenario_done_cb)
        self.app.proxy_manager.disconnect_by_func(self.__asset_transcoding_progress_cb)
        self.app.proxy_manager.disconnect_by_func(self.__proxy_error_cb)
        self.app.proxy_manager.disconnect_by_func(self.__asset_transcoding_cancelled_cb)
        self.app.proxy_manager.disconnect_by_func(self.__proxy_ready_cb)

    def save(self, ges_timeline, uri, formatter_asset, overwrite):
        for container_profile in self.list_encoding_profiles():
            if not isinstance(container_profile, GstPbutils.EncodingContainerProfile):
                self.warning("%s is not an EncodingContainerProfile", container_profile)
                continue

            for profile in container_profile.get_profiles():
                preset_factory_name = profile.get_preset_name()
                preset = Gst.ElementFactory.make(preset_factory_name, None)
                if not isinstance(preset, Gst.Preset):
                    self.warning("Element %s does not implement Gst.Preset. Cannot save"
                                 "its rendering settings", preset)
                    continue

                if profile.get_type_nick() == "video":
                    cache = self._vcodecsettings_cache
                elif profile.get_type_nick() == "audio":
                    cache = self._acodecsettings_cache
                else:
                    self.warning("Unrecognized profile type for profile %s", profile)
                    continue
                cache_key = (container_profile, preset_factory_name)
                if cache_key not in cache:
                    continue

                current_preset = profile.get_preset()
                if current_preset and current_preset.startswith("encoder_settings_"):
                    current_preset = None

                # The settings for the current GstPbutils.EncodingProfile.
                settings = cache[cache_key]
                # The name of the Gst.Preset storing the settings.
                preset_name = "encoder_settings_%s" % uuid.uuid4().hex
                # The project has three GstPbutils.EncodingProfile,
                # for the container, for the video, for the audio.
                # Each of them keeps the encoding settings in a Gst.Preset
                # saved externally in a separate file.
                # When the project is loaded, the encoding settings are loaded
                # automatically.
                profile.set_preset(preset_name)

                # The original preset name is also important.
                preset.set_meta(preset_name, "pitivi::OriginalPreset", current_preset)

                # Store the current GstPbutils.EncodingProfile's settings
                # in the Gst.Preset.
                for prop, value in settings.items():
                    preset.set_property(prop, value)

                # Serialize the GstPbutils.EncodingProfile's settings
                # from the cache into e.g.
                # $XDG_DATA_HOME/gstreamer-1.0/presets/GstX264Enc.prs
                # for x264enc presets.
                res = preset.save_preset(preset_name)
                assert res

        return GES.Project.save(self, ges_timeline, uri, formatter_asset, overwrite)

    def use_proxies_for_assets(self, assets, scaled=False):
        proxy_manager = self.app.proxy_manager
        originals = []
        for asset in assets:
            if proxy_manager.asset_can_be_proxied(asset, scaled):
                target = asset.get_proxy_target()
                uri = proxy_manager.get_proxy_uri(asset, scaled=scaled)
                if target and target.props.id == uri:
                    self.info("Missing proxy needs to be recreated after cancelling"
                              " its recreation")
                    target.unproxy(asset)

                originals.append(asset)

        if originals:
            with self.app.action_log.started("Proxying assets"):
                for asset in originals:
                    action = AssetProxiedIntention(asset, self, self.app.proxy_manager)
                    self.app.action_log.push(action)
                    self._prepare_asset_processing(asset)
                    asset.force_proxying = True
                    proxy_manager.add_job(asset, scaled)

    def disable_proxies_for_assets(self, assets, delete_proxy_file=False, hq_proxy=True):
        proxy_manager = self.app.proxy_manager
        for asset in assets:
            if proxy_manager.is_proxy_asset(asset):
                proxy_target = asset.get_proxy_target()
                # The asset is a proxy for the proxy_target original asset.
                self.debug("Stop proxying %s", proxy_target.props.id)
                proxy_target.set_proxy(None)
                if proxy_manager.is_scaled_proxy(asset) \
                        and not proxy_manager.is_asset_format_well_supported(proxy_target) \
                        and hq_proxy:
                    # The original asset is unsupported, and the user prefers
                    # to edit with HQ proxies instead of scaled proxies.
                    self.use_proxies_for_assets([proxy_target])
                self.remove_asset(asset)
                proxy_target.force_proxying = False
                if delete_proxy_file:
                    os.remove(Gst.uri_get_location(asset.props.id))
            else:
                # The asset is an original which is not being proxied.
                proxy_manager.cancel_job(asset)

    def regenerate_scaled_proxies(self):
        assets = self.list_assets(GES.Extractable)
        scaled_proxies = []
        scaled_proxy_targets = []

        for asset in assets:
            if self.app.proxy_manager.is_scaled_proxy(asset):
                scaled_proxies.append(asset)
                scaled_proxy_targets.append(asset.get_proxy_target())

        self.disable_proxies_for_assets(scaled_proxies, delete_proxy_file=True,
                                        hq_proxy=False)
        self.use_proxies_for_assets(scaled_proxy_targets, scaled=True)

    def _commit(self):
        """Logs the operation and commits.

        To be used as a replacement for the GES.Timeline.commit method, allowing
        to scenarialize the action in the scenarios.
        """
        self.app.write_action("commit")
        GES.Timeline.commit(self.ges_timeline)

    def create_timeline(self):
        """Loads the project's timeline."""
        try:
            # The project is loaded from the file in this call.
            self.ges_timeline = self.extract()
        except GLib.Error as e:
            self.warning("Failed to extract the timeline: %s", e)
            self.ges_timeline = None

        if self.ges_timeline is None:
            return False

        self.ges_timeline.commit = self._commit
        self.pipeline = Pipeline(self.app)
        if not self.pipeline.set_timeline(self.ges_timeline):
            self.warning("Failed to set the pipeline's timeline: %s", self.ges_timeline)
            return False

        if self.ges_timeline.get_marker_list("markers") is None:
            self.ges_timeline.set_marker_list("markers", GES.MarkerList.new())

        return True

    def update_restriction_caps(self):
        """Syncs the project settings from the render profile.

        The project settings reside in the restriction caps of the audio and
        video tracks of the timeline. They are updated to match the render
        profile stored as the first (and only) encoding profile of the project.
        """
        videocaps = Gst.Caps.new_empty_simple("video/x-raw")
        # Get the height/width without rendering settings applied
        width, height = self.get_video_width_and_height()
        videocaps.set_value("width", width)
        videocaps.set_value("height", height)
        videocaps.set_value("framerate", self.videorate)
        videocaps.set_value("pixel-aspect-ratio", Gst.Fraction(1, 1))

        audiocaps = Gst.Caps.new_empty_simple("audio/x-raw")
        audiocaps.set_value("rate", self.audiorate)
        audiocaps.set_value("channels", self.audiochannels)

        for track in self.ges_timeline.get_tracks():
            if isinstance(track, GES.VideoTrack):
                track.set_restriction_caps(videocaps)
            elif isinstance(track, GES.AudioTrack):
                track.set_restriction_caps(audiocaps)

        if self.app:
            self.app.write_action(
                "set-track-restriction-caps",
                caps=videocaps.to_string(),
                track_type=GES.TrackType.VIDEO.value_nicks[0])

            self.app.write_action(
                "set-track-restriction-caps",
                caps=audiocaps.to_string(),
                track_type=GES.TrackType.AUDIO.value_nicks[0])

        self.pipeline.commit_timeline()

    def add_uris(self, uris):
        """Adds assets asynchronously.

        Args:
            uris (List[str]): The URIs of the assets.
        """
        with self.app.action_log.started("assets-addition"):
            for uri in uris:
                if self.create_asset(quote_uri(uri), GES.UriClip):
                    # The asset was not already part of the project.
                    action = AssetAddedIntention(self, uri)
                    self.app.action_log.push(action)

    def assets_for_uris(self, uris):
        assets = []
        for uri in uris:
            asset = self.get_asset(uri, GES.UriClip)
            if not asset:
                return None
            assets.append(asset)
        return assets

    def list_sources(self):
        return self.list_assets(GES.UriClip)

    def get_project_id(self):
        project_id = self.get_string("pitivi::project-id")
        if not project_id:
            project_id = uuid.uuid4().hex
            self.set_string("pitivi::project-id", project_id)
            self.log("Assigned new project id %s", project_id)
        return project_id

    def release(self):
        res = 0

        if self.pipeline:
            self.pipeline.release()

        if self.runner:
            res = self.runner.printf()

        if self.runner:
            self.runner = None
            self.monitor = None

        self.pipeline = None
        self.ges_timeline = None

        return res

    def set_modification_state(self, state):
        if not self.loaded:
            return

        self._dirty = state
        if state:
            self.emit('project-changed')

    def has_unsaved_modifications(self):
        return self._dirty

    def get_dar(self):
        return Gst.Fraction(self.videowidth, self.videoheight)

    def get_video_width_and_height(self, render=False):
        """Returns the video width and height as a tuple.

        Args:
            render (bool): Whether to apply self.render_scale to the returned
                values.
        """
        if render:
            if not self._has_rendering_values:
                return (self.videowidth * self.render_scale / 100,
                        self.videoheight * self.render_scale / 100)
            else:
                return self.videowidth, self.videoheight

        if self._has_rendering_values:
            return (self.videowidth / self.render_scale * 100,
                    self.videoheight / self.render_scale * 100)

        return self.videowidth, self.videoheight

    def get_video_caps(self, render=False):
        """Gets the caps corresponding to the video settings.

        Returns:
            Gst.Caps: The video settings caps.
        """
        videowidth, videoheight = self.get_video_width_and_height(render=render)
        vstr = "width=%d,height=%d,pixel-aspect-ratio=1/1,framerate=%d/%d" % (
            videowidth, videoheight,
            self.videorate.num, self.videorate.denom)
        caps_str = "video/x-raw,%s" % (vstr)
        video_caps = Gst.caps_from_string(caps_str)
        return video_caps

    def get_audio_caps(self):
        """Gets the caps corresponding to the audio settings.

        Returns:
            Gst.Caps: The audio settings caps.
        """
        astr = "rate=%d,channels=%d" % (self.audiorate, self.audiochannels)
        caps_str = "audio/x-raw,%s" % (astr)
        audio_caps = Gst.caps_from_string(caps_str)
        return audio_caps

    def set_audio_properties(self, nbchanns=-1, rate=-1):
        """Sets the number of audio channels and the rate."""
        # pylint: disable=consider-using-in
        self.info("%d x %dHz %dbits", nbchanns, rate)
        if nbchanns != -1 and nbchanns != self.audiochannels:
            self.audiochannels = nbchanns
        if rate != -1 and rate != self.audiorate:
            self.audiorate = rate

    def set_encoders(self, muxer="", vencoder="", aencoder=""):
        """Sets the video and audio encoders and the muxer."""
        # pylint: disable=consider-using-in
        if muxer != "" and muxer != self.muxer:
            self.muxer = muxer
        if vencoder != "" and vencoder != self.vencoder:
            self.vencoder = vencoder
        if aencoder != "" and aencoder != self.aencoder:
            self.aencoder = aencoder

    @property
    def containersettings(self):
        cache_key = (self.container_profile, self.muxer)
        return self._containersettings_cache.setdefault(cache_key, {})

    @containersettings.setter
    def containersettings(self, value):
        cache_key = (self.container_profile, self.muxer)
        self._containersettings_cache[cache_key] = value

    @property
    def vcodecsettings(self):
        cache_key = (self.container_profile, self.vencoder)
        return self._vcodecsettings_cache.setdefault(cache_key, {})

    @vcodecsettings.setter
    def vcodecsettings(self, value):
        cache_key = (self.container_profile, self.vencoder)
        self._vcodecsettings_cache[cache_key] = value

    @property
    def acodecsettings(self):
        cache_key = (self.container_profile, self.aencoder)
        return self._acodecsettings_cache.setdefault(cache_key, {})

    @acodecsettings.setter
    def acodecsettings(self, value):
        cache_key = (self.container_profile, self.aencoder)
        self._acodecsettings_cache[cache_key] = value

    # ------------------------------------------ #
    # Private methods                            #
    # ------------------------------------------ #

    def _ensure_tracks(self):
        if self.ges_timeline is None:
            self.warning("Can't ensure tracks if no timeline set")
            return

        track_types = [track.get_property("track-type")
                       for track in self.ges_timeline.get_tracks()]

        if GES.TrackType.VIDEO not in track_types:
            self.ges_timeline.add_track(GES.VideoTrack.new())
        if GES.TrackType.AUDIO not in track_types:
            self.ges_timeline.add_track(GES.AudioTrack.new())

    def _ensure_layer(self):
        if self.ges_timeline is None:
            self.warning("Can't ensure tracks if no timeline set")
            return
        if not self.ges_timeline.get_layers():
            self.ges_timeline.append_layer()

    def _ensure_restrictions(self, profile, defaults, ref_restrictions=None,
                             prev_vals=None):
        """Make sure restriction values defined in @defaults are set on @profile.

        Attributes:
            profile (Gst.EncodingProfile): The Gst.EncodingProfile to use
            defaults (dict): A key value dict to use to set restriction defaults
            ref_restrictions (Gst.Caps): Reuse values from those caps instead
                                         of @values if available.

        """
        if isinstance(profile, GstPbutils.EncodingAudioProfile):
            facttype = Gst.ELEMENT_FACTORY_TYPE_AUDIO_ENCODER
        else:
            facttype = Gst.ELEMENT_FACTORY_TYPE_VIDEO_ENCODER

        ebin = Gst.ElementFactory.make("encodebin", None)
        ebin.props.profile = profile

        encoder = None
        for element in ebin.iterate_recurse():
            if element.get_factory().list_is_type(facttype):
                encoder = element
                break

        if not encoder:
            self.error("element '%s' not available for profile %s", profile.get_preset(), profile)
            return False

        encoder_sinkcaps = encoder.sinkpads[0].get_pad_template().get_caps().copy()
        self.debug("%s - Ensuring %s\n  defaults: %s\n  ref_restrictions: %s\n  prev_vals: %s)",
                   encoder, encoder_sinkcaps, defaults, ref_restrictions,
                   prev_vals)
        restriction = fixate_caps_with_default_values(encoder_sinkcaps,
                                                      ref_restrictions,
                                                      defaults,
                                                      prev_vals)
        assert restriction
        preset_name = encoder.get_factory().get_name()
        profile.set_restriction(restriction)
        profile.set_preset_name(preset_name)

        self.info("Fully set restriction: %s", profile.get_restriction().to_string())
        return True

    def _ensure_video_restrictions(self, profile=None):
        defaults = {
            "width": 720,
            "height": 576,
            "framerate": Gst.Fraction(25, 1),
            "pixel-aspect-ratio": Gst.Fraction(1, 1),
        }

        prev_vals = None
        if self.video_profile:
            prev_vals = self.video_profile.get_restriction().copy()

        if not profile:
            profile = self.video_profile
            ref_restrictions = Gst.Caps("video/x-raw")
        else:
            ref_restrictions = profile.get_restriction()

        return self._ensure_restrictions(profile, defaults, ref_restrictions, prev_vals)

    def _ensure_audio_restrictions(self, profile=None):
        ref_restrictions = None
        if not profile:
            profile = self.audio_profile
        else:
            ref_restrictions = profile.get_restriction()

        defaults = {"channels": Gst.IntRange(range(1, 2147483647)),
                    "rate": Gst.IntRange(range(8000, GLib.MAXINT))}
        prev_vals = None
        if self.audio_profile:
            prev_vals = self.audio_profile.get_restriction().copy()

        return self._ensure_restrictions(profile, defaults, ref_restrictions, prev_vals)

    def _maybe_init_settings_from_asset(self, asset):
        """Updates the project settings to match the specified asset.

        Args:
            asset (GES.UriClipAsset): The asset to copy the settings from.
        """
        if not (self._has_default_video_settings or
                self._has_default_audio_settings):
            # Both audio and video settings have been set already by the user.
            return
        if not isinstance(asset, GES.UriClipAsset):
            # We are only interested in actual files, not in titles, for example.
            return

        emit = False
        info = asset.get_info()
        video_streams = info.get_video_streams()
        if video_streams and self._has_default_video_settings:
            video = video_streams[0]
            if not video.is_image():
                videowidth = video.get_natural_width()
                videoheight = video.get_natural_height()
                videorate = self.videorate
                if video.get_framerate_num() > 0:
                    # The asset has a non-variable framerate.
                    videorate = Gst.Fraction(video.get_framerate_num(),
                                             video.get_framerate_denom())
                self.set_video_properties(videowidth, videoheight, videorate)
                self._has_default_video_settings = False
                emit = True
        audio_streams = info.get_audio_streams()
        if audio_streams and self._has_default_audio_settings:
            audio = audio_streams[0]
            if audio.get_sample_rate() >= DECENT_AUDIORATE:
                self.audiochannels = audio.get_channels()
                self.audiorate = audio.get_sample_rate()
                self._has_default_audio_settings = False
                emit = True
        if emit:
            self.emit("settings-set-from-imported-asset", asset)

    def get_element_factory_name(self, profile):
        """Finds a factory for an element compatible with the specified profile.

        Args:
            profile (GstPbutils.EncodingProfile): A muxer, video or audio
                profile.

        Returns:
            str: The name of the factory which can produce the required
            element, or None.
        """
        if profile.get_preset_name():
            return profile.get_preset_name()

        factories = Project.__factories_compatible_with_profile(profile)
        if not factories:
            return None

        preset = profile.get_preset()
        if not preset:
            # The element does not need to support a specific preset.
            # Return the compatible factory with the highest rank.
            return factories[0].get_name()
        else:
            # Make sure that if a #Gst.Preset is set we find an
            # element that can handle that preset.
            for factory in factories:
                elem = factory.create()
                if isinstance(elem, Gst.Preset):
                    if elem.load_preset(preset):
                        return factory.get_name()

            self.error("Could not find any element with preset %s", preset)
            return None

    @staticmethod
    def __factories_compatible_with_profile(profile):
        """Finds factories of the same type as the specified profile.

        Args:
            profile (GstPbutils.EncodingProfile): A muxer, video or audio
                profile.

        Returns:
            list[Gst.ElementFactory]: The element factories producing elements
                of the same type as the specified profile.
        """
        element_factories = []
        if isinstance(profile, GstPbutils.EncodingContainerProfile):
            element_factories = Encoders().muxers
        elif isinstance(profile, GstPbutils.EncodingVideoProfile):
            element_factories = Encoders().vencoders
        elif isinstance(profile, GstPbutils.EncodingAudioProfile):
            element_factories = Encoders().aencoders
        else:
            raise ValueError("Profile type not handled: %s" % profile)

        factories = Gst.ElementFactory.list_filter(element_factories,
                                                   Gst.Caps(profile.get_format()),
                                                   Gst.PadDirection.SRC,
                                                   False)

        factories.sort(key=lambda x: - x.get_rank())

        return factories



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2009, Brandon Lewis <brandon_lewis@berkeley.edu>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Dialog for locating a missing asset."""
import os
from gettext import gettext as _

from gi.repository import Gtk

from pitivi.mediafilespreviewer import PreviewWidget
from pitivi.medialibrary import AssetThumbnail
from pitivi.utils.loggable import Loggable
from pitivi.utils.ui import beautify_missing_asset
from pitivi.utils.ui import PADDING
from pitivi.utils.ui import SPACING


class MissingAssetDialog(Gtk.Dialog, Loggable):
    """Dialog for locating a missing asset using Gtk.FileChooserWidget.

    Attributes:
        app (Pitivi): The app.
        asset (GES.UriClipAsset): The missing asset.
        uri (str): Last known URI of the missing asset.
    """

    def __init__(self, app, asset, uri):
        Gtk.Dialog.__init__(self)
        Loggable.__init__(self)

        self.set_title(_("Locate missing file..."))
        self.set_modal(True)
        self.add_buttons(_("Cancel"), Gtk.ResponseType.CANCEL,
                         _("Open"), Gtk.ResponseType.OK)
        self.set_border_width(SPACING * 2)
        self.get_content_area().set_spacing(SPACING)
        self.set_transient_for(app.gui)
        self.set_default_response(Gtk.ResponseType.OK)

        # This box will contain widgets with details about the missing file.
        vbox = Gtk.Box()
        vbox.set_orientation(Gtk.Orientation.VERTICAL)

        label_start = Gtk.Label()
        label_start.set_markup(_("The following file could not be found:"))
        label_start.set_xalign(0)
        vbox.pack_start(label_start, False, False, 0)

        hbox = Gtk.Box()
        hbox.set_orientation(Gtk.Orientation.HORIZONTAL)
        hbox.set_margin_top(PADDING)
        hbox.set_spacing(PADDING * 2)

        label_asset_info = Gtk.Label()
        label_asset_info.set_markup(beautify_missing_asset(asset))
        label_asset_info.set_xalign(0)
        label_asset_info.set_yalign(0)
        hbox.pack_start(label_asset_info, False, False, 0)

        unused_small_thumb, large_thumb = AssetThumbnail.get_thumbnails_from_xdg_cache(uri)
        if large_thumb:
            self.debug("A thumbnail file was found for %s", uri)
            thumbnail = Gtk.Image.new_from_pixbuf(large_thumb)
            hbox.pack_end(thumbnail, False, False, 0)

        vbox.pack_start(hbox, False, False, 0)

        label_end = Gtk.Label()
        label_end.set_markup(_("Please specify its new location:"))
        label_end.set_xalign(0)
        label_end.set_margin_top(PADDING)
        vbox.pack_start(label_end, False, False, 0)

        self.get_content_area().pack_start(vbox, False, False, 0)
        vbox.show_all()

        self._chooser = self.__setup_file_chooser(uri, app.settings)
        self.get_content_area().pack_start(self._chooser, True, True, 0)
        self._chooser.show()

        # If the window is too big, the window manager will resize it so that
        # it fits on the screen.
        self.set_default_size(1024, 1000)

    @staticmethod
    def __setup_file_chooser(uri, settings):
        chooser = Gtk.FileChooserWidget(action=Gtk.FileChooserAction.OPEN)
        chooser.set_select_multiple(False)
        previewer = PreviewWidget(settings, discover_sync=True)
        chooser.set_preview_widget(previewer)
        chooser.set_use_preview_label(False)
        chooser.connect("update-preview", previewer.update_preview_cb)
        chooser.set_current_folder(settings.lastProjectFolder)
        # Use a Gtk FileFilter to only show files with the same extension
        # Note that splitext gives us the extension with the ".", no need to
        # add it inside the filter string.
        unused_filename, extension = os.path.splitext(uri)
        filter_ = Gtk.FileFilter()
        # Translators: this is a format filter in a filechooser. Ex: "AVI files"
        filter_.set_name(_("%s files") % extension)
        filter_.add_pattern("*%s" % extension.lower())
        filter_.add_pattern("*%s" % extension.upper())
        default = Gtk.FileFilter()
        default.set_name(_("All files"))
        default.add_pattern("*")
        chooser.add_filter(filter_)
        chooser.add_filter(default)
        return chooser

    def get_new_uri(self):
        """Returns new URI of the missing asset, if provided by the user."""
        response = self.run()
        if response == Gtk.ResponseType.OK:
            self.log("User chose a new URI for the missing file")
        return self._chooser.get_uri()
